/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	26-Jul-2005
 */
#ifndef __DATA_USRBIN_H
#define __DATA_USRBIN_H

#include <assert.h>
#include "math/matrix4.h"

#define USRBIN_RUNTIT	80
#define	USRBIN_RUNTIM	32
#define USRBIN_TITUSB	10

// ================================== Usrbin ===================================
/*
 * If the standard system is used
 *	X = horizontal
 *	Y = vertical
 *	Z = horizontal-beam direction
 * Then the transformation that has to be performed are
 *	X -> -X		[X]  Y   Z	Size: -1.0
 *	Y ->  Z		 X   Y  [Z]	Size:  1.0
 *	Z ->  Y		 X  [Y]  Z	Size:  1.0
 */
class Usrbin {
public:
	enum class Type : int {
		XYZ_point      =  0,
		XYZ            = 10,

		RPhiZ_point    =  1,
		RPhiZ          = 11,

		region_point   =  2,
		region         = 12,

		Xsym_point     =  3,
		Xsym           = 13,

		Ysym_point     =  4,
		Ysym           = 14,

		Zsym_point     =  5,
		Zsym           = 15,

		XYZsym_point   =  6,
		XYZsym         = 16,

		RPhiZsym_point =  7,
		RPhiZsym       = 17,

		special_point  =  8,
		special        = 18
	}; // Type

private:
	int	_detector;		// detector loaded
	Type	_type;			// type of usrbin
	int	_score;			// quantity scored
public:
	int	nx,    ny,    nz;
	double	xlow,  ylow,  zlow;
	double	xhigh, yhigh, zhigh;
	double	dx,    dy,    dz;
//	Point	offset;
	double	xofs,  yofs,  zofs;	// offset of usrbin
	double	x0,  y0;		// for R-Z-Phi binning
	double	min, max;
	char	runtitle[USRBIN_RUNTIT+1],
		runtime[USRBIN_RUNTIM+1];
	char	name[USRBIN_TITUSB+1];

	int	lntzer;			// not zero flag
	float	bkusbn;
	float	b2usbn;
	float	tcusbn;

	float	weipri;
	int	ncase;
	int	mcase;
	int	nbatch;
	std::string	filename;

private:
	double	_norm;
	bool	_logscale;		// is logarithmic histogram
	bool	_checker;
	bool	_hasMatrix;		// if rotation matrix is set
	Matrix4	_matrix;

	int	nynx;
	time_t	_mtime;

	float*	data;			// FIXME convert to double
//	float*	error;
//	double	 _total;

static	const char*	_typeStr[];

public:
	Usrbin()			: data(NULL) { cleanup(); }
	Usrbin(const char* fn, int det)	: data(NULL) { load(fn,det); }
	~Usrbin()	{ cleanup(); }

	void	cleanup();
	bool	load(const char* fn, int det);
	bool	save(const char* fn=NULL);

	/* create usrbin */
	void	create(int det, Type atype, int ascore,
			const double axlow, const double axhigh, const int anx,
			const double aylow, const double ayhigh, const int any,
			const double azlow, const double azhigh, const int anz);
	void	checker(const Type atype,
			double axlow, double axhigh, const int anx,
			double aylow, double ayhigh, const int any,
			double azlow, double azhigh, const int anz);
	bool	checker()	const	{ return _checker; }

	/* get/set */
	int	detector()	const	{ return _detector; }
	Type	type()	const	{ return _type; }
	int	score()		const	{ return _score; }

	bool	hasData()	const	{ return data != NULL; }

	double	xmin()		const	{
			if (_type==Type::XYZsym_point || _type==Type::XYZsym ||
			    _type==Type::Xsym_point   || _type==Type::Xsym)
				return -xhigh;
			else
				return xlow;
		}
	double	ymin()		const	{
			if (_type==Type::XYZsym_point || _type==Type::XYZsym ||
			    _type==Type::Ysym_point   || _type==Type::Ysym)
				return -yhigh;
			else
				return ylow;
		}
	double	zmin()		const	{
			if (_type==Type::XYZsym_point   || _type==Type::XYZsym ||
			    _type==Type::Zsym_point     || _type==Type::Zsym   ||
			    _type==Type::RPhiZsym_point || _type==Type::RPhiZsym)
				return -zhigh;
			else
				return zlow;
		}
	double	xmax()			const	{ return xhigh; }
	double	ymax()			const	{ return yhigh; }
	double	zmax()			const	{ return zhigh; }

	void	norm(const double n);
	double	norm()			const	{ return _logscale? exp10(_norm) : _norm; }
	void	normalize(const double n);

	bool	convert(bool tolog);
	bool	logscale()		const	{ return _logscale; }

	// WARNING no transformation is applied!!!
	int	usrbini(const double x)	const	{ return Int((x - xlow)/dx); }
	int	usrbinj(const double y)	const	{ return Int((y - ylow)/dy); }
	int	usrbink(const double z)	const	{ return Int((z - zlow)/dz); }

	double	usrbinx(const int i)	const	{ return xlow + i*dx; }
	double	usrbiny(const int j)	const	{ return ylow + j*dy; }
	double	usrbinz(const int k)	const	{ return zlow + k*dz; }

	double	usrbincx(const int i)	const	{ return xlow + i*dx + dx/2.0; }
	double	usrbincy(const int j)	const	{ return ylow + j*dy + dy/2.0; }
	double	usrbincz(const int k)	const	{ return zlow + k*dz + dz/2.0; }

	// with transformation
	bool	xyz2ijk(double x, double y, double z, int* i, int* j, int* k) const;

	//   getData returns the normalized data as stored in the memory
	//   WARNING if _logscale is set, it returns the logarithm of the data
	double	getData(const int ptr) const		{ return _normValue(data[ptr]); }
	double	getData(int i, int j, int k) const	{ return _normValue(data[k*nynx + j*nx + i]); }

	bool	getData(double x, double y, double z, double* value) const;
	bool	getData(const int region, double* value) const;
	bool	getDataRPhiZ(double r, double phi, double z, double* value) const;

	/** @return real normalized value for cell indexed ptr */
	double	get(const int ptr) const {
				if (_logscale)
					return exp10(getData(ptr));
				else
					return getData(ptr);
			}

	/** @return real normalized value at location i,j,k */
	double	get(int i, int j, int k) const	{
				if (_logscale)
					return exp10(getData(i,j,k));
				else
					return getData(i,j,k);
			}

	/** @return real normalized value at location */
	bool	get(double x, double y, double z, double* value) const {
				bool ok = getData(x,y,z,value);
				if (ok && _logscale)
					*value = exp10(*value);
				return ok;
			}

	/** @return real normalized value at location given by point P */
	double	get(const Point& P) const {
				double value;
				if (get(P.x, P.y, P.z, &value)) {
					return value;
				} else
					return 0.0;
			}

	/** @return real normalized value for the region */
	bool	get(const int region, double* value) const {
				bool ok = getData(region, value);
				if (ok && _logscale)
					*value = exp10(*value);
				return ok;
			}

	/** set a value to a cell */
	void	set(int i, int j, int k, float value) {
				assert(!_logscale && !_checker && data);
				data[k*nynx + j*nx + i] = value;
			}

	/** increase cell value by amount v */
	void	add(int i, int j, int k, float value) {
				assert(!_logscale && !_checker && data);
				data[k*nynx + j*nx + i] += value;
				/*_total += value;*/
			}
	void	add(double x, double y, double z, float value) {
				assert(!_logscale && !_checker && data);
				int i,j,k;
				if (xyz2ijk(x,y,z, &i, &j, &k)) add(i,j,k,value);
			}
	double	add(Point A, Point B, double value=1.0, bool tracklength=true);

	/** @return integral on a straight line */
	double	integrate(Point A, Point B, double t) const;

	void	offset(const Point& ofs);
	void	scanMinMax();

	// transformation matrix operations
	void	clearMatrix()			{ _hasMatrix = false; }
	bool	hasMatrix()		const	{ return _hasMatrix; }
	void	matrix(const Matrix4& M);
const	Matrix4& matrix()		const	{ return _matrix; }

private:
	// returned normalized value
	double	_normValue(double value) const {
				if (_logscale)
					return value + _norm;
				else
					return value * _norm;
			}

	double	intersectXYZ(Point& P, const Vector& D) const;
	void	_xyzStepInit(double pos, double dir, double low, double high, double* t, double* dt, int* step, int* out) const;
}; // Usrbin

#endif
