/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "mesh.h"
#include "boxbody.h"

using std::cout;
using std::endl;

// ================================= GBOXBody ==================================
/** set */
void GBOXBody::set(const double x, const double y, const double z,
		   const double ux, const double uy, const double uz,
		   const double vx, const double vy, const double vz,
		   const double wx, const double wy, const double wz)
{
	assert(type() == Type::BOX);

	P.set( x,  y,  z);
	Z.set(ux, uy, uz);
	X.set(vx, vy, vz);
	Y.set(wx, wy, wz);

	// outer corner
	Po = P + X + Y + Z;

	xlen = X.normalize();
	ylen = Y.normalize();
	zlen = Z.normalize();

	checkOrthogonal();
} // set

/** set */
void GBOXBody::set(const double xmin, const double xmax,
		   const double ymin, const double ymax,
		   const double zmin, const double zmax)
{
	assert(type() == Type::RPP);

	P.set( xmin, ymin, zmin);
	Po.set(xmax, ymax, zmax);

	X = Vector::Xo;
	Y = Vector::Yo;
	Z = Vector::Zo;

	xlen = Po.x - P.x;
	ylen = Po.y - P.y;
	zlen = Po.z - P.z;
} // set

/** set */
void GBOXBody::set(double *what)
{
	resetBBox();
	round0(what, nwhats(), SMALL);
	if (type() == Type::RPP) {
		set(what[0], what[1],
		    what[2], what[3],
		    what[4], what[5]);
	} else {
		set(what[0], what[ 1], what[ 2],
		    what[3], what[ 4], what[ 5],
		    what[6], what[ 7], what[ 8],
		    what[9], what[10], what[11]);
	}
} // set

/* --- get --- */
int GBOXBody::get(double *what) const
{
	if (type() == Type::RPP) {
		what[0] = Min(P.x, Po.x);
		what[1] = Max(P.x, Po.x);
		what[2] = Min(P.y, Po.y);
		what[3] = Max(P.y, Po.y);
		what[4] = Min(P.z, Po.z);
		what[5] = Max(P.z, Po.z);
		return 6;
	} else {
		what[ 0] = P.x;
		what[ 1] = P.y;
		what[ 2] = P.z;
		what[ 3] = zlen*Z.x;
		what[ 4] = zlen*Z.y;
		what[ 5] = zlen*Z.z;
		what[ 6] = xlen*X.x;
		what[ 7] = xlen*X.y;
		what[ 8] = xlen*X.z;
		what[ 9] = ylen*Y.x;
		what[10] = ylen*Y.y;
		what[11] = ylen*Y.z;
		return 12;
	}
} // get

/** position
 * @param x,y,z	new position to set
 */
void GBOXBody::position(const Point& r)
{
	GBody::position(r);
	Po = P + xlen*X + ylen*Y + zlen*Z;
} // position

/** rotate body around an axis */
void GBOXBody::rotate(const double angle, const Vector& axis)
{
	GBody::rotate(angle,axis);
	Po = P + xlen*X + ylen*Y + zlen*Z;

	// Check type
	Vector::Direction ax = X.direction(SMALL);
	Vector::Direction ay = Y.direction(SMALL);
	Vector::Direction az = Z.direction(SMALL);

	if (ax!=Vector::Direction::Unknown and
	    ay!=Vector::Direction::Unknown and
	    az!=Vector::Direction::Unknown) {
		_type = Type::RPP;

		X = Vector::Xo;
		Y = Vector::Yo;
		Z = Vector::Zo;

		if (P.x > Po.x) Swap(P.x, Po.x);
		if (P.y > Po.y) Swap(P.y, Po.y);
		if (P.z > Po.z) Swap(P.z, Po.z);
		xlen = Po.x - P.x;
		ylen = Po.y - P.y;
		zlen = Po.z - P.z;
	} else
		_type = Type::BOX;
} // rotate

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GBOXBody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	Point tr = _hasMatrix? _invMatrix*r : r;
	tr -= SP;
	if (item==1) { // Lower X-plane
		double xp = tr*SX;
		P = SP + xp*SX;
		xlen = sxlen - xp;
		if (xlen<0.0) {
			xlen = -xlen;
			P = SP + sxlen*SX;
		}
	} else
	if (item==2) {	// Lower Y-plane
		double yp = tr*SY;
		P = SP + yp*SY;
		ylen = sylen - yp;
		if (ylen<0.0) {
			ylen = -ylen;
			P = SP + sylen*SY;
		}
	} else
	if (item==3) { // Lower Z-plane
		double zp = tr*SZ;
		P = SP + zp*SZ;
		zlen = szlen - zp;
		if (zlen<0.0) {
			zlen = -zlen;
			P = SP + szlen*SZ;
		}
	} else {
		// Combined moves
		//    16 = YZ plane (X-perpendicular)
		//    17 = XZ plane (Y-perpendicular)
		//    18 = XY plane (Z-perpendicular)

		bool found = false;
		P = SP;
		if (item==4 || item==17 || item==18) { // Upper X-plane
			found = true;
			xlen = tr*SX;
			if (xlen<0.0) {
				xlen = -xlen;
				P -= xlen*SX;
			}
		}

		if (item==5 || item==16 || item==18) { // Upper Y-plane
			found = true;
			ylen = tr*SY;
			if (ylen<0.0) {
				ylen = -ylen;
				P -= ylen*SY;
			}
		}

		if (item==6 || item==16 || item==17) { // Upper Z-plane
			found = true;
			zlen = tr*SZ;
			if (zlen<0.0) {
				zlen = -zlen;
				P -= zlen*SZ;
			}
		}

		if (!found)
			GBody::move(item, r, w);
	}
	Po = P + xlen*X + ylen*Y + zlen*Z;
} // move

/** createQuads */
void GBOXBody::createQuads()
{
	GBody::createQuads();
	if (type() == Type::RPP) {
		//        Cx   Cy   Cz   C
		addQuad(-1.0, 0.0, 0.0, P.x);
		addQuad( 0.0,-1.0, 0.0, P.y);
		addQuad( 0.0, 0.0,-1.0, P.z);

		addQuad( 1.0, 0.0, 0.0,-Po.x);
		addQuad( 0.0, 1.0, 0.0,-Po.y);
		addQuad( 0.0, 0.0, 1.0,-Po.z);
	} else {
		// first 3 planes will be
		//        Cx     Cy     Cz      C
		addQuad(-X.x,-X.y,-X.z, X*P);
		addQuad(-Y.x,-Y.y,-Y.z, Y*P);
		addQuad(-Z.x,-Z.y,-Z.z, Z*P);

		// last 3 planes will be
		addQuad( X.x, X.y, X.z,-X*Po);
		addQuad( Y.x, Y.y, Y.z,-Y*Po);
		addQuad( Z.x, Z.y, Z.z,-Z*Po);
	}
} // createQuads

/* createMesh */
void GBOXBody::createMesh()
{
	GBody::createMesh();
	if (_mesh->isEmpty())
		_mesh->createParallelepiped(P, xlen*X, ylen*Y, zlen*Z);
	else {
		// update vertices
		_mesh->vertex(0) = P;			// 0
		_mesh->vertex(1) = P + xlen*X;		// 1
		_mesh->vertex(2) = P + xlen*X + ylen*Y;	// 2
		_mesh->vertex(3) = P          + ylen*Y;	// 3

		_mesh->vertex(4) = P                   + zlen*Z; // 4
		_mesh->vertex(5) = P + xlen*X          + zlen*Z; // 5
		_mesh->vertex(6) = P + xlen*X + ylen*Y + zlen*Z; // 6
		_mesh->vertex(7) = P          + ylen*Y + zlen*Z; // 7
		_mesh->calcBbox();
	}
} // GBOXBody

/** update bounding box of body */
void GBOXBody::updateBBox() const
{
	Point  p = GBody::position();
	Vector x = xlen * vectorX();
	Vector y = ylen * vectorY();
	Vector z = zlen * vectorZ();

	// Add the 8 vertices
	_bbox.add(p);
	_bbox.add(p + x);
	_bbox.add(p + x + y);
	_bbox.add(p     + y);
	_bbox.add(p         + z);
	_bbox.add(p + x     + z);
	_bbox.add(p + x + y + z);
	_bbox.add(p     + y + z);
} // updateBBox

#ifdef OBBOX
/** update oriented bounding box of body */
void GBOXBody::updateOBB(bool in) const
{
	OBBox& bb = _obbox[(int)in];
	bb.P = GBody::position();
	bb.X = vectorX();
	bb.Y = vectorY();
	bb.Z = vectorZ();
	bb.low()  = Point(0.0,  0.0,  0.0);
	bb.high() = Point(xlen, ylen, zlen);
} // updateOBB
#endif

/**
 * @param body	other body to check location against
 * @return location of this body with respect to another body
 */
GBody::Location GBOXBody::_locationWrt(const GBody* body) const
{
#if 0
	if (body->type() == Type::SPH) {
		Location bodyloc = Location::AinB;
		//bodysize holds half of the size of the BOX in each of its dimensions
		Vector bodysize = Vector(vectorXlen()/2, vectorYlen()/2, vectorZlen()/2);
		// Rsph locates the SPHERE center w.r.t the center of the BOX
		Vector Rsph = GBody::position() - (body->position() + bodysize);
		//d holds the closest boundary (vertex, edge) in the direction of the SPHERE position
		Vector d;
		d.x = Rsph.x ? bodysize.x * Rsph.x / Abs(Rsph.x) : 0.0;
		d.y = Rsph.y ? bodysize.y * Rsph.y / Abs(Rsph.y) : 0.0;
		d.z = Rsph.z ? bodysize.z * Rsph.z / Abs(Rsph.z) : 0.0;
		//if the SPHERE is outside the BOX...
		if (Rsph.length() >  d.length()) {
			//dist1 holds distance from SPHERE center to BOX's closest boundary (vertex, edge)
			double dist1 = (Rsph - d).length();
			//dist2 holds distance from SPHERE center to BOX's farthest boundary (vertex, edge)
			double dist2 = (Rsph + d).length();

			if (((const GSPHBody*)body)->R() < dist1)
				bodyloc = Location::OUTSIDE;
			else
			if (((const GSPHBody*)body)->R() < dist2)
				bodyloc = Location::OVERLAP;
			else
				bodyloc = Location::AinB; //check, should be BOX in SPHERE
		} else { //...else if the sphere is inside the BOX
			//not implemented
			bodyloc = Location::OVERLAP;
			float Radio = ((const GSPHBody*)body)->R();
			if (Rsph.x + Radio < xlen)
				if (Rsph.y + Radio < ylen)
					if (Rsph.z + Radio < zlen)
						bodyloc = Location::BinA; //check, should be SPHERE in BOX
		}
	} //else
	// if (body->type() == Type::XCC || body->type() == Type::YCC ||
	//     body->type() == Type::ZCC) {
	//	// Minimum distance between center and axes
	//	double dist = PointLineDistance(position(),
	//		body->position(), body->vectorZ());

	//	if (dist - ((const GInfEllCylBody*)body)->R() >  R())
	//		return LOCATION_OUTSIDE;
	//	else
	//		return LOCATION_OVERLAP;
	// } else
	// if (body->type() == Type::BOX || body->type() == Type::RPP) {

	//	}
		// return bodyloc;
	// }//--------------------------------------------added by Leonel---------
#endif
	return GBody::_locationWrt(body);
} // _locationWrt

// ================================= GWEDBody ==================================
/** set */
void GWEDBody::set(const double x, const double y, const double z,
		   const double ux, const double uy, const double uz,
		   const double vx, const double vy, const double vz,
		   const double wx, const double wy, const double wz)
{
	P.set( x,  y,  z);
	X.set(ux, uy, uz);
	Y.set(vx, vy, vz);
	Z.set(wx, wy, wz);

	// Upper point
	Po = P + Z;

	// normalize vectors
	xlen = X.normalize();
	ylen = Y.normalize();
	zlen = Z.normalize();

	checkOrthogonal();
} // set

/** set */
void GWEDBody::set(double *what)
{
	resetBBox();
	round0(what, nwhats(), SMALL);
	set(what[0], what[ 1], what[ 2],
	    what[3], what[ 4], what[ 5],
	    what[6], what[ 7], what[ 8],
	    what[9], what[10], what[11]);
} // set

/* --- get --- */
int GWEDBody::get(double *what) const
{
	what[ 0] = P.x;
	what[ 1] = P.y;
	what[ 2] = P.z;
	what[ 3] = xlen*X.x;
	what[ 4] = xlen*X.y;
	what[ 5] = xlen*X.z;
	what[ 6] = ylen*Y.x;
	what[ 7] = ylen*Y.y;
	what[ 8] = ylen*Y.z;
	what[ 9] = zlen*Z.x;
	what[10] = zlen*Z.y;
	what[11] = zlen*Z.z;
	return 12;
} // get

/** @return normal of the cutting plane */
Vector	GWEDBody::N() const
{
	// normal of the cutting plane
	Vector n = (ylen*Y - xlen*X) ^ Z;
	n.normalize();
	return n;
} // N

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GWEDBody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	Point tr = _hasMatrix? _invMatrix*r : r;
	tr -= SP;
	switch (item) {
		case 1: {	// Lower X-plane
				double xp = tr*SX;
				P = SP + xp*SX;
				xlen = sxlen - xp;
				if (xlen<0.0) {
					X = -SX;
					xlen = -xlen;
				} else
					X = SX;
			}
			break;

		case 2: {	// Lower Y-plane
				double yp = tr*SY;
				P = SP + yp*SY;
				ylen = sylen - yp;
				if (ylen<0.0) {
					Y = -SY;
					ylen = -ylen;
				} else
					Y = SY;
			}
			break;

		case 3: {	// Lower Z-plane
				double zp = tr*SZ;
				P = SP + zp*SZ;
				zlen = szlen - zp;
				if (zlen<0.0) {
					Z = -SZ;
					zlen = -zlen;
				} else
					Z = SZ;
			}
			break;

		case 4:		// Upper Z-plane
			zlen = tr*SZ;
			if (zlen<0.0) {
				zlen = -zlen;
				P = SP - zlen*SZ;
			}
			break;

		case 5: {	// cutting plane
				// normal from the saved variables
				Vector n = (sylen*SY - sxlen*SX) ^ SZ;
				n.normalize();
				double d = tr*n;
				Point R = SP + d*n;
				if (d<0.0) {
					X = -SX;
					Y = -SY;
					d = -d;
				} else {
					X = SX;
					Y = SY;
				}
				double xr = X*(R-P);
				if (!Eq0(xr,SMALL3)) {
					xlen = Sq(d) / xr;
					ylen = Sq(d) / (Y*(R-P));
				}
			}
			break;

		default:
			GBody::move(item, r, w);
	}
	Po = P + zlen*Z;
} // move

/** createQuads */
void GWEDBody::createQuads()
{
	GBody::createQuads();
	// first 3 planes will be
	//        Cx     Cy     Cz      C
	addQuad(-X.x,-X.y,-X.z, X*P);
	addQuad(-Y.x,-Y.y,-Y.z, Y*P);
	addQuad(-Z.x,-Z.y,-Z.z, Z*P);

	// upper Z plane
	addQuad( Z.x, Z.y, Z.z,-Z*Po);

	// normal of the cutting plane
	Vector n = N();

	// cutting plane
	double d = -n * (P+xlen*X);
	if (n*X < 0.0) {
		n = -n;
		d = -d;
	}
	addQuad( n.x, n.y, n.z, d);
} // createQuads

/* createMesh */
void GWEDBody::createMesh()
{
	GBody::createMesh();

	bool first = _mesh->isEmpty();
	if (first) _mesh->allocateVertices(6);

	_mesh->vertex(0) = P;
	_mesh->vertex(1) = P  + xlen*X;
	_mesh->vertex(2) = P  + ylen*Y;
	_mesh->vertex(3) = Po;
	_mesh->vertex(4) = Po + xlen*X;
	_mesh->vertex(5) = Po + ylen*Y;
	_mesh->calcBbox();

	if (first) {
		_mesh->add(0, 2, 1, true,  true,  true );
		_mesh->add(3, 4, 5, true,  true,  true );

		_mesh->add(0, 4, 3, false, true,  true );
		_mesh->add(0, 1, 4, true,  true,  false);

		_mesh->add(0, 3, 5, true,  true,  false);
		_mesh->add(0, 5, 2, false, true,  true );

		_mesh->add(1, 2, 5, true,  true,  false);
		_mesh->add(1, 5, 4, false, true,  true );

		_mesh->process();
		assert(_mesh->isClosed());
		assert(_mesh->isOrientable());
#if _DEBUG>2
		cout << endl;
		cout << "WED Mesh ";
		cout << " isClosed=" << _mesh->isClosed();
		cout << " isOrientable=" << _mesh->isOrientable() << endl;
		cout << "WED volume=" << _mesh->volume() << endl;
#endif
	}
} // createMesh

/** update bounding box of body */
void GWEDBody::updateBBox() const
{
	Point  p = position();
	Vector x = xlen * vectorX();
	Vector y = ylen * vectorY();
	Vector z = zlen * vectorZ();

	// Add the 6 vertices
	_bbox.add(p);
	_bbox.add(p + x);
	_bbox.add(p + y);
	_bbox.add(p     + z);
	_bbox.add(p + x + z);
	_bbox.add(p + y + z);
} // updateBBox

#ifdef OBBOX
/** update oriented bounding box of body */
void GWEDBody::updateOBB(bool in) const
{
	OBBox& bb = _obbox[(int)in];
	bb.P = GBody::position();
	bb.X = vectorX();
	bb.Y = vectorY();
	bb.Z = vectorZ();
	bb.low() = Point(0.0, 0.0, 0.0);
	if (!in)
		bb.high() = Point(xlen, ylen, zlen);
	else
		bb.high() = Point(xlen/2.0, ylen/2.0, zlen);
} // updateOBB
#endif
