/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 *
 * Author:	Christian.Theis@cern.ch
 *
 * Date: 20-Nov-2020
 */
#ifndef BOUNDING_VOL_HIERARCHY_H
#define BOUNDING_VOL_HIERARCHY_H

#include <vector>

#include "geo.h"
#include "math/bbox.h"
#include "math/vector.h"
#include "geometry/engine.h"
#include "geometry/vregion.h"

class Ray;
class VZone;

int QSplit( VRegion** ppNodeList, int NrOfNodes, double PivotVal, int Axis);
void MergeBoundingBoxes( const Vector& Box1Min, const Vector& Box1Max, const Vector& Box2Min, const Vector& Box2Max, Vector& NewBoxMin, Vector& NewBoxMax );

class CBoundingVolHierarchy
{
public:
  enum  { cXAxis = 0, cYAxis = 1, cZAxis = 2 };

  CBoundingVolHierarchy( GeometryEngine* pEngine, std::vector<VRegion*> RegionList, size_t  NrOfNodes, bool bVerbose = false, int LogLevel = 0 );
  CBoundingVolHierarchy( GeometryEngine* pEngine, VRegion* pGeoRegion );
  CBoundingVolHierarchy( GeometryEngine* pEngine, CBoundingVolHierarchy* pNode1, CBoundingVolHierarchy* pNode2 );
  CBoundingVolHierarchy( GeometryEngine* pEngine, CBoundingVolHierarchy* pNode1, CBoundingVolHierarchy* pNode2, const BBox& CurBBox );
  virtual ~CBoundingVolHierarchy( void ) = default;

  // copy ctor
  CBoundingVolHierarchy( const CBoundingVolHierarchy& /*rhs*/ ) = default;
  CBoundingVolHierarchy& operator=( const CBoundingVolHierarchy& /*rhs*/ ) = default;

  // build a branch starting at ppStartNode using NrOfSubBranchNodes which will all be enclosed by one new bounding volume
  virtual CBoundingVolHierarchy*  BuildBranch( VRegion** ppStartNode, size_t NrOfSubBranchNodes, int Axis = cXAxis);

  // check if the passed point is inside the BVH and return the respective zone or a nullptr
  virtual VZone*    inside( const Vector& CurPoint, double eps = SMALL3D4 );

  // intersect a ray with the BVH and return nullptr or the intersected zone
  VZone* intersectRay( Ray* pCurRay );

  // determine the number of nodes from a subset (given by ppStartNode + NrOfNodes) that have a centroid coordinate < PivotVal with respect to a specific axis.
  static  size_t    QSplit( VRegion** ppStartNode, size_t NrOfNodes, double PivotVal, int Axis);

  unsigned int      GetTotalNrOfLeaves() const { return m_NrOfLeaves; };

  size_t            GetNrOfRegions() const { return m_NodeList.size(); };

  void              PrintTree();

  // return bounding box
  const BBox&       bbox() { return m_BoundingBox; }

protected:

  // internal function to recursively check if the passed point is contained
  VZone* Contains( const Vector& CurPoint, double eps );

  // output current leaf
  void              PrintLeaf( int Level = 0 );
	void              Indent( int Level ) {
		for( int i = 0; i < Level * 3; ++i )
			std::cerr << " ";
	}

  // optimized intersection that avoids pre-calculating constants like the inverse direction
	VZone* intersectRay( Ray* pCurRay, const Vector& Pos, const Vector& InvDir, double& tmin, double tmax );


public:

protected:
  CBoundingVolHierarchy*        m_pLeftLeaf;
  CBoundingVolHierarchy*        m_pRightLeaf;
  VRegion*                      m_pGeoRegion;

  BBox            m_BoundingBox;
  unsigned int    m_NrOfLeaves;

  bool            m_bVerbose;
  unsigned int    m_LogLevel;

  std::vector<VRegion*> m_NodeList;

  GeometryEngine* m_pEngine;
};

#endif
