/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	10-Oct-2018
 */

#ifndef __GEOMETRY_INFCYLGBODY_H
#define __GEOMETRY_INFCYLGBODY_H

#include "gbody.h"

// =============================== GCylinderBase ===============================
// Base class of all cylinder like bodies RCC, REC, TRC, ... to create the mesh
class GCylinderBase : public GBody {
public:
	static int N_CYLINDER;

public:
	GCylinderBase(const std::string& aname, const Type atype) : GBody(aname, atype) {}
protected:
	void	makeConeQuads();
	// mesh functions
	void	createEllConeMesh(const double Rbx, const double Rby,
				  const double Rax, const double Ray,
				  const bool isinfinite=false);
}; // class GCylinderBase

// ================================= GRCCBody ==================================
class GRCCBody : public GCylinderBase {
public:
	GRCCBody(const std::string& aname) : GCylinderBase(aname, Type::RCC) {}

	double	R()	const		{ return xlen; }
	void	R(const double r)	{ xlen = ylen = r; }

	void	set(const double x,  const double y,  const double z,
		    const double hx, const double hy, const double hz,
		    const double r);
	void	set(const Vector& pos, const Vector& h, const double r)
			{ set(pos.x, pos.y, pos.z, h.x, h.y, h.z, r); }
	virtual	void	set(double*);
	virtual	int	get(double*) const;
	virtual int	nwhats()	const	{ return 7; }
	virtual	int	nodes()		const	{ return 2; }
	virtual void	move(int item, const Point& r, const Vector& w);

// FIXME
//	virtual	double	distance(const double x, const double y, const double z) const;

	// Dimension labels and show flags
	// @return NULL not to show anything otherwise label
	virtual	const char*	showX()	const { return "R"; }
	virtual const char*	showY()	const { return NULL; }
	virtual const char*	showZ()	const { return "H"; }

protected:
	virtual	void	createQuads();
	virtual void	createMesh()	{ createEllConeMesh(xlen,ylen,xlen,ylen); }
	virtual	void	updateBBox() const;
#ifdef OBBOX
	virtual	void	updateOBB(bool in=false) const;
#endif
}; // GRCCBody

// ================================= GRECBody ==================================
class GRECBody : public GCylinderBase {
public:
	GRECBody(const std::string& aname) : GCylinderBase(aname, Type::REC) {}
	void	set(const double x,  const double y,  const double z,
		    const double wx, const double wy, const double wz,
		    const double ux, const double uy, const double uz,
		    const double vx, const double vy, const double vz);
	void	set(const Vector& pos, const Vector& w,
		    const Vector& u,   const Vector& v)
			{ set(pos.x, pos.y, pos.z,
			      w.x, w.y, w.z,
			      u.x, u.y, u.z,
			      v.x, v.y, v.z); }
	virtual	void	set(double*);
	virtual	int	get(double*) const;
	virtual int	nwhats()	const	{ return 12; }
	virtual	int	nodes()		const	{ return 4; }
	virtual	int	closest(const Point& r, const double d, const Vector& w) const;
	virtual void	move(int item, const Point& r, const Vector& w);

// FIXME
//	virtual	double	distance(const double x, const double y, const double z) const;

	// Dimension labels and show flags
	// @return NULL not to show anything otherwise label
	virtual	const char*	showX()	const { return "Rx"; }
	virtual const char*	showY()	const { return "Ry"; }
	virtual const char*	showZ()	const { return "H"; }

protected:
	virtual	void	createQuads();
	virtual void	createMesh()	{ createEllConeMesh(xlen,ylen,xlen,ylen); }
	virtual	void	updateBBox() const;
#ifdef OBBOX
	virtual	void	updateOBB(bool in=false) const;
#endif
}; // GRECBody

// ================================= GTRCBody ==================================
class GTRCBody : public GCylinderBase {
private:
	double	h2;
public:
	GTRCBody(const std::string& aname) : GCylinderBase(aname, Type::TRC) {}

	double	Rb()	const		{ return xlen; }
	double	Ra()	const		{ return ylen; }
	void	Rb(const double r)	{ xlen = r; }
	void	Ra(const double r)	{ ylen = r; }

	void	set(const double x,  const double y,  const double z,
		    const double hx, const double hy, const double hz,
		    const double rb, const double ra);
	void	set(const Vector& pos, const Vector& h,
		    const double rb, const double ra)
			{ set(pos.x, pos.y, pos.z, h.x, h.y, h.z, rb, ra); }
	virtual	void	set(double*);
	virtual	int	get(double*) const;
	virtual int	nwhats()	const	{ return 8; }
	virtual	int	nodes()		const	{ return 2; }
	virtual	int	closest(const Point& r, const double d, const Vector& w) const;
	virtual void	move(int item, const Point& r, const Vector& w);

// FIXME
//	virtual	double	distance(const double x, const double y, const double z) const;

	// Dimension labels and show flags
	// @return NULL not to show anything otherwise label
	virtual	const char*	showX()	const { return "Rb"; }
	virtual const char*	showY()	const { return "Ra"; }
	virtual const char*	showZ()	const { return "H"; }

protected:
	virtual	void	createQuads();
	virtual void	createMesh()	{ createEllConeMesh(Rb(),Rb(),Ra(),Ra()); }
	virtual	void	updateBBox() const;
#ifdef OBBOX
	virtual	void	updateOBB(bool in=false) const;
#endif
}; // GTRCBody

// ============================== GInfEllCylBody ===============================
class GInfEllCylBody : public GCylinderBase {
public:
	GInfEllCylBody(const std::string& aname, const Type atype)
		: GCylinderBase(aname, atype)
			{ zlen = infinite; }

	Vector	direction() const;

	double	R()	const		{ return xlen; };
	double	Rx()	const		{ return xlen; };
	double	Ry()	const		{ return ylen; };
	void	R(const double r)	{ xlen = ylen = r; }
	void	Rx(const double r)	{ xlen = r; }
	void	Ry(const double r)	{ ylen = r; }

	void	set(const double a, const double b, const double ra, const double rb=0.0);

	virtual	void	set(double*);
	virtual	int	get(double*) const;
	virtual int	whats() const {
				return In(type(), Type::XCC, Type::YCC, Type::ZCC)? 3 : 4;
			}
	virtual	int	closest(const Point& r, const double d, const Vector& w) const;
	virtual void	move(int item, const Point& r, const Vector& w);
	virtual	void	rotate(const double angle, const Vector& axis);

// FIXME
//	virtual	double	distance(const double x, const double y, const double z) const;

	// Dimension labels and show flags
	// @return NULL not to show anything otherwise label
	virtual	const char*	showX()	const {
				switch (type()) {
					case Type::XCC:
					case Type::YCC:
					case Type::ZCC: return "R";
					case Type::XEC: return "Ry";
					case Type::YEC: return "Rz";
					case Type::ZEC: return "Rx";
					default: return NULL;
				}
			}
	virtual const char*	showY()	const {
				switch (type()) {
					case Type::XCC:
					case Type::YCC:
					case Type::ZCC: return NULL;
					case Type::XEC: return "Rz";
					case Type::YEC: return "Rx";
					case Type::ZEC: return "Ry";
					default: return NULL;
				}
			}
	virtual const char*	showZ()	const { return NULL; }

protected:
	virtual	void	createQuads();
	virtual void	createMesh() { createEllConeMesh(xlen,ylen,xlen,ylen,true); }
	virtual	void	updateBBox() const;
#ifdef OBBOX
	virtual	void	updateOBB(bool in=false) const;
#endif
	virtual	Location _locationWrt(const GBody *body) const;
}; // GInfEllCylBody

#endif
