/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	20-Apr-2020
 */

#include "volume.h"

#include "gregion.h"
#include "geometry.h"
#include "math/bit.h"
#include "math/bbox.h"
#include "math/randsampler.h"

//using namespace flair;
using std::cout;
using std::endl;

// =============================== VolumeWorker ================================
#ifdef THREAD
/** operator() */
void VolumeWorker::operator()()
{
	VolumeFeeder* f = static_cast<VolumeFeeder*>(feeder());
const	VolumeJob* job  = f->job;
const	GRegion* region = job->region;
const	Point& low    = job->bbox.low();
const	Point& center = job->bbox.centroid();
	double radius = (center - low).length();

	while (idx<to) {
		Point  pos;
		Vector dir;
		double rnd1 = CBaseSampler::VanDerCorput(idx, random.seed());
		double rnd2 = CBaseSampler::Sobol2(idx, random.seed());
		random.sflood(&pos, &dir, rnd1, rnd2);
		pos *= radius;
		pos += center;
		double tl = 2*radius;	// very crude!!!!
		++cache;
		bdx += region->boundaryCrossings(pos, dir, &tl, cache);
		tracklength += tl;
		idx++;
	}
} // operator

// =============================== VolumeFeeder ================================
/** allocate workers if the requested number of threads changed */
void VolumeFeeder::allocate()
{
	if (nworkers != pool.nthreads()) {
		if (workers) delete [] workers;
		nworkers = pool.nthreads();
		workers = new VolumeWorker[nworkers];
		for (int i=0; i<nworkers; i++)
			workers[i].feeder(this);
	}
} // allocate

/** reset feeder */
void VolumeFeeder::reset(const VolumeJob* j, int nsamples, int nstep, Random& random)
{
	pool.reset();
	allocate();

	job     = j;
	idx     = 0;
	step    = nstep;
	samples = nsamples;

	// initialize workers
	for (int i=0; i<nworkers; i++)
		workers[i].reset(random.integer(), job->geometry->nBodies());
} // reset

/** feed */
ThreadPoolWorker* VolumeFeeder::feed(int threadId)
{
	// Check for ending condition
	if (idx>=samples) return nullptr;

	// update worker
	workers[threadId].idx = idx;
	workers[threadId].to  = idx = Min(idx + step, samples);	// next value

	return (ThreadPoolWorker*)&workers[threadId];
} // feed
#endif

// ================================== VolumeJob ==================================
/** Calculate the surface of the current region using Monte Carlo integration
 * @param region	region to find volume of
 * @param samples (I/O)	I:maximum number of samples to run, O:current status
 * @param vol (O)	return volume during calculation
 * @param err (O)	return error of volume calculation
 * @return surface area of region
 */

/** initialize job
 * @return true if everything is ok
 */
bool VolumeJob::init(const Geometry* geo, const GRegion* reg, int samples, int threads)
{
	geometry = geo;
	region = reg;
	volume = area = error = 0.0;
	if (region) {
		bbox  = region->bbox();
		if (!bbox.isValid())
			return false;
	} else {
		bbox.reset();
		return false;
	}
	center = bbox.centroid();
	radius = (bbox.high() - center).length();
	std::cout << "center=" << center << " radius=" << radius << std::endl;

	run_samples = samples;
	if (!isPowerOf2(run_samples))
		run_samples = roundUpPow2(run_samples);
	nsamples = 0;

	initThreads(threads);

	return true;
} // init

/** run area job, either using the threadpool or in a single thread */
void VolumeJob::run()
{
	startRun();

	long bdx = 0;
	double tracklength = 0;
	Random random;
	nsamples = 0;

	double piR2 = PI * Sq(radius);
#ifdef THREAD
	if (nthreads() > 1) {
		VolumeFeeder feeder(threadpool);
		for (int n=32; n<=run_samples; n<<=1) {
			feeder.reset(this, n, 1024, random);
			if (threadpool.execute(&feeder)) break;
			if (threadpool.isStopped()) break;
			bdx = feeder.bdx();
			tracklength = feeder.tracklength();
			// Calculate temporarily for reporting
			nsamples = n;
			volume = piR2 * tracklength / nsamples;
			// For isotropic flux the current is twice as the flux
			area   = 2.0 * piR2 * bdx / nsamples;
			// Poisson stats var=n
			error  = 1.0 / sqrt((double)nsamples);
			DUMP(cout << nsamples
				<< " / " << run_samples
				<< " Volume=" << volume
				<< " Area=" << area
				<< " +/ " << error << endl);
		}
	} else
#endif
	{
		Cache<CBody3D> cache(geometry->nBodies());
		int bp = 1<<5;	// breakpoint to show progress
		for (int n=0; n<run_samples; n++) {
			Point  pos;
			Vector dir;
			double rnd1 = CBaseSampler::VanDerCorput(n, random.seed());
			double rnd2 = CBaseSampler::Sobol2(n, random.seed());
			random.sflood(&pos, &dir, rnd1, rnd2);
			pos *= radius;
			pos += center;

			// count how many intersections we have with the current region when going
			// from pos along dir
			++cache;
			double tl = 2*radius;	// very crude!!!!
			bdx += region->boundaryCrossings(pos, dir, &tl, cache);
			tracklength += tl;
			//bdx += region->nrOfIntersections(pos, dir, cache);

			// Calculate temporarily for reporting
			if (++nsamples > bp) {
				volume = piR2 * tracklength / nsamples;
				// For isotropic flux the current is twice as the flux
				area   = 2.0 * piR2 * bdx / nsamples;
				// Poisson stats var=n
				error  = 1.0 / sqrt((double)nsamples);
				cout << nsamples << ":  Volume=" << volume
						<< " Area=" << area
						<< " +/ " << error << endl;
				bp <<= 1;
			}
		}
	}
	//assert(bdx == feeder.bdx());
	if (bdx) {
		volume = piR2 * tracklength / nsamples;
		// For isotropic flux the current is twice as the flux
		area   = 2.0 * piR2 * bdx / nsamples;
		// Poisson stats var=n
		error  = 1.0 / sqrt((double)nsamples);
		DUMP(cout << "Region: " << region->name()
			<< ", bsphere (" << center
			<< "), bdx: " << bdx
			<< ", volume: " << volume
			<< ", area: " << area
			<< "+/-" << error << endl);
	}
	endRun();
} // run

/** spawn */
void VolumeJob::spawn()
{
	pthread_create(&thread, NULL,
		(void*(*)(void*))VolumeJob::runInThread, this);
	//stop();
	//thread = std::thread(&VolumeJob::run, this);
	//thread.detach();
} // spawn

/** runInThread */
void* VolumeJob::runInThread(VolumeJob* job)
{
	job->run();
	job->thread = 0;
	return NULL;
} // runInThread
