/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	01-Apr-2011
 */

#include "vzone.h"
#include "kernel.h"
#include "engine.h"
#include "tool/stack.h"

using std::cout;
using std::endl;

#if defined(_DUMP) && defined(STRONG)
	#define RNAME	"R1"
	#define XCHK	3.35680660661
	#define YCHK	-0.866272672673
	#define ACHK	0.1
//	#define CHECK(X) if (!strcmp(region->name(),RNAME)) X
//	#define CHECK(X) if (!strcmp(region->name(),RNAME) && Abs(x-XCHK)<ACHK && Abs(y-YCHK)<ACHK) X
//	#define CHECK(X) if (Abs(x-XCHK)<ACHK && Abs(y-YCHK)<ACHK) X
	#define CHECK(X)	X
#else
	#define CHECK(X)
#endif

// A pair for stack operations inside and body
struct InsideBody {
	CBody*	body;
	bool	inside;
};

/** init */
void VZone::init(GZone* z, VRegion *r)
{
	_zone       = z;
	_region     = r;
	_generation = -1;
	_location   = false;
	_in         = false;
} // init

/** vBody */
inline VBody* VZone::vbody(const GBody* body) const
{
	return _region->_kernel->getBody(body);
} // vbody

/** cbody */
inline CBody* VZone::cbody(GeometryEngine* engine, const GBody* body)
{
	return engine->getBody(body);
} // cbody

/** updateLocation
 *
 * find the location of the Zone on the Z=0 plane and within the viewport
 * coordinates
 *
 * FIXME: Improvement on STD expressions we can ignore directly the zones
 *        with a negative result, and remove the bodies
 *        Probably add a reference count to the bodies!
 *
 * We have 3 possible locations outside(0) / inside(1) / unknown(?:2)
 *		  a + b
 *	0 + 0 = 0	? + 0 = 0
 *	0 + 1 = 0	? + 1 = ?
 *	1 + 0 = 0	0 + ? = 0
 *	1 + 1 = 1	1 + ? = ?
 *
 *		  a - b
 *	0 - 0 = 0	? - 0 = ?
 *	0 - 1 = 0	? - 1 = 0
 *	1 - 0 = 1	0 - ? = 0
 *	1 - 1 = 0	1 - ? = ?
 *
 *		  a | b
 *	0 | 0 = 0	? | 0 = ?
 *	0 | 1 = 1	? | 1 = 1
 *	1 | 0 = 1	0 | ? = ?
 *	1 | 1 = 1	1 | ? = 1
 */
void VZone::updateLocation()
{
#if 0 // Find location using the obb
	int viewLoc = obbox()->intersectWithPlane(
			_region->viewer->view.matrix,
			_region->viewer->view.minX,
			_region->viewer->view.maxX,
			_region->viewer->view.minY,
			_region->viewer->view.maxY);

	_location = false;
	if (viewLoc == 1) {
		_location = true;
		_in = true;
		return;
	} else if (viewLoc == 0) {
		_location = true;
		_in = false;
		return;
	}

#else
	_location = false;
	if (size()==0) {
		_location = true;
		_in = false;
		return;
	}
	if (std()) {
		int  prod = 1;	// result of product
		int  i=0;
		while (i<size()) {
			const GBody *body = gexpr(i++);
			if (body == &GBody::tminus) break;
			int bloc = (int)vbody(body)->location;
			if (prod<=1 && bloc<=1)
				prod = prod && bloc;
			else
			if (prod==0 || bloc==0)
				prod = 0;
			else
				prod = 2;
			if (prod==0) {
				_location = true;
				_in = false;
				return;
			}
		}
		while (i<size()) {
			const GBody *body = gexpr(i++);
			// FIXME
			int bloc = (int)vbody(body)->location;
			if (prod<=1 && bloc<=1)
				prod = prod && !bloc;
			else
			if (prod==0 || bloc==1)
				prod = 0;
			else
				prod = 2;
			if (prod==0) {
				_location = true;
				_in = false;
				return;
			}
		}
		// New zone or end...
		if (prod==1) {
			_location = true;
			_in = true;
		}
	} else { // RPN
		Stack<int, GZone::ZONE_STACK_SIZE> stack;
		for (int i=0; i<size(); i++) {
			const GBody *body = gexpr(i);
			if (body == &GBody::tplus) {
				int b = stack.pop();
				if (stack()<=1 && b<=1)
					stack.set((int)(stack() and b));
				else
				if (stack()==0 or b==0)
					stack.set(0);
				else
					stack.set(2);
			} else
			if (body == &GBody::tminus) {
				int b = stack.pop();
				if (stack()<=1 && b<=1)
					stack.set((int)(stack() and !b));
				if (stack()==0 or b==1)
					stack.set(0);
				else
					stack.set(2);
			} else
			if (body == &GBody::tunion) {
				int b = stack.pop();
				if (stack()<=1 and b<=1)
					stack.set((int)(stack() or b));
				if (stack()==1 or b==1)
					stack.set(1);
				else
					stack.set(2);
			} else
			if (body == &GBody::tuniverse)
				stack.push(1);
			else
				stack.push((int)vbody(body)->location);
		}
		if (stack.ptr() != 0) {	// Error recovery
			_location = true;
			_in = false;
			return;
		}
		if (stack() <= 1) {
			_location = true;
			_in = stack();
		}
	}
#endif
#ifdef STRONG
//	if (_location)
//		cout << "Region: " << region->name()
//		     << "\tlocation=" << _location << "\tin=" << _in << endl;
#endif
} // updateLocation

/** _inside2D */
bool VZone::_inside2D(GeometryEngine* engine,
		      const double  x, const double  y, const double  z,
		      const double dx, const double dy, const double dz) const
{
	if (size()==0) return false;
	CHECK( cout << endl << "VZone: " << *this
			<< " pos= " << x << " " << y << " " << z
			<< " dir= " << dx << " " << dy << " " << dz << endl);
	if (std()) {
		int  i=0;
		// first plus terms (if any)
		while (i<size()) {
			const GBody *body = gexpr(i++);
			if (body == &GBody::tminus) break;
#ifdef EXPERIMENTAL
			CHECK(int checked=cbody(engine, body)->isChecked());
			CHECK(cout << "   + " << body->name()
				     << " in["<<checked<<"]="
				     << cbody(engine, body)->inside2D(engine,x,y,z,dx,dy,dz) << endl);
			if (!cbody(engine, body)->inside2D(engine,x,y,z,dx,dy,dz)) return false;
#else
			if (!cbody(engine, body)->inside2D(x,y,z,dx,dy,dz)) return false;
#endif
		}
		// then negative terms (if any)
		while (i<size()) {
			const GBody *body = gexpr(i++);
#ifdef EXPERIMENTAL
			CHECK(int checked=cbody(engine, body)->isChecked());
			CHECK(cout << "   - " << body->name()
				   << " in["<<checked<<"]="
				   <<  cbody(engine, body)->inside2D(engine,x,y,z,dx,dy,dz) << endl);
			if (cbody(engine, body)->inside2D(engine,x,y,z,dx,dy,dz)) return false;
#else
			if (cbody(engine, body)->inside2D(x,y,z,dx,dy,dz)) return false;
#endif
		}
		CHECK(cout << "   | prod=1" << endl);
		return true;
	} else { // RPN
		bool b;
		Stack<bool, GZone::ZONE_STACK_SIZE> stack;

		for (int i=0; i<size(); i++) {
			const GBody *body = gexpr(i);
			if (body == &GBody::tplus) {
				b = stack.pop();
				stack.set(stack() and b);
				CHECK(cout << "  +  " << stack() << endl);
			} else
			if (body == &GBody::tminus) {
				b = stack.pop();
				stack.set(stack() and !b);
				CHECK(cout << "  -  " << stack() << endl);
			} else
			if (body == &GBody::tunion) {
				b = stack.pop();
				stack.set(stack() or b);
				CHECK(cout << "  |  " << stack() << endl);
			} else
			if (body == &GBody::tuniverse) {
				stack.push(true);
				CHECK(cout << "  @  " << stack() << endl);
			} else {
				CHECK(int checked=cbody(engine, body)->isChecked());
#ifdef EXPERIMENTAL
				stack.push(cbody(engine, body)->inside2D(engine,x,y,z,dx,dy,dz));
#else
				stack.push(cbody(engine, body)->inside2D(x,y,z,dx,dy,dz));
#endif
				CHECK(cout << " --> "<<body->name()
					   << " in["<<checked<<"]="
					   << "=" << stack() << ":\t";
					for (int j=0; j<stack.size(); j++) cout << stack[j] << ' ';
					cout << endl);
			}
		}
		CHECK(cout << region()->name() << "=" << stack() << endl << endl);
		assert(stack.ptr()==0);
		return stack();
	}
} // _inside2D

/** _inside */
// FIXME duplication should be merged with the inside(x,y,z, dx,dy,dz, t=0)
bool VZone::inside(GeometryEngine* engine,
		      const double  x, const double  y, const double  z,
		      const double dx, const double dy, const double dz) const
{
	if (size()==0) return false;
	CHECK( cout << endl << "VZone: " << *this
			<< " pos= " << x << " " << y << " " << z
			<< " dir= " << dx << " " << dy << " " << dz << endl);
	if (std()) {
		int  i=0;
		// first plus terms (if any)
		while (i<size()) {
			const GBody *body = gexpr(i++);
			if (body == &GBody::tminus) break;
			if (!cbody(engine, body)->inside(x,y,z, dx,dy,dz)) return false;
		}
		// then negative terms (if any)
		while (i<size()) {
			const GBody *body = gexpr(i++);
			if (cbody(engine, body)->inside(x,y,z, dx,dy,dz)) return false;
		}
		CHECK(cout << "   | prod=1" << endl);
		return true;
	} else { // RPN
		//Stack<InsideBody, GZone::ZONE_STACK_SIZE> stack;
		bool stack[GZone::ZONE_STACK_SIZE], b;
		int  sp = -1;	/* stack pointer */
		// Perform a lazy(late) calculation only when it is needed
		CBody* bstack[GZone::ZONE_STACK_SIZE];	// NULL if already pushed to stack,
							// else needs evaluation of inside
		for (int i=0; i<size(); i++) {
			const GBody *body = gexpr(i);
			// a+b = a && b
			if (body == &GBody::tplus) {
				if (!sp) return false;
				assert(sp>0);
				// check if any of the two is known and false
				if ((bstack[sp]==NULL   and !stack[sp]) or		// b
				    (bstack[sp-1]==NULL and !stack[sp-1])) {		// a
					// (a or b)=false => a+b = false
					sp--;
					stack[sp]  = false;
					bstack[sp] = NULL;
				} else {
					// retrieve b
					if (bstack[sp])
						b = bstack[sp--]->inside(x,y,z, dx,dy,dz);
					else
						b = stack[sp--];
					if (b) {
						// b = true => a+b = a
						if (bstack[sp]) {
							stack[sp]  = bstack[sp]->inside(x,y,z, dx,dy,dz);
							bstack[sp] = NULL;
						} // else already there no need to do anything
					} else {
						// b = false => a+b = false
						stack[sp]  = false;
						bstack[sp] = NULL;
					}
				}
				CHECK(cout << "  +  " << stack[sp] << endl);
			} else
			// a-b = a && !b
			if (body == &GBody::tminus) {
				if (!sp) return false;
				assert(sp>0);
				// check if any of the two is known and false
				if ((bstack[sp]==NULL   and  stack[sp]) or		//!b
				    (bstack[sp-1]==NULL and !stack[sp-1])) {		// a
					// (a or !b)=false => a-b = false
					sp--;
					stack[sp]  = false;
					bstack[sp] = NULL;
				} else {
					// retrieve b
					if (bstack[sp])
						b = bstack[sp--]->inside(x,y,z, dx,dy,dz);
					else
						b = stack[sp--];
					if (!b) {
						// b=false => a-b = a
						if (bstack[sp]) {
							stack[sp]  = bstack[sp]->inside(x,y,z, dx,dy,dz);
							bstack[sp] = NULL;
						} // else already there no need to do anything
					} else {
						// b=true => a-b = false
						stack[sp]  = false;
						bstack[sp] = NULL;
					}
				}
				CHECK(cout << "  -  " << stack[sp] << endl);
			} else
			// a | b
			if (body == &GBody::tunion) {
				if (!sp) return false;
				assert(sp>0);
				// check if any of the two is known and true
				if ((bstack[sp]==NULL   and  stack[sp]) or		// b
				    (bstack[sp-1]==NULL and  stack[sp-1])) {		// a
					// (a or b)=true => a|b = true
					sp--;
					stack[sp]  = true;
					bstack[sp] = NULL;
				} else {
					// retrieve b
					if (bstack[sp])
						b = bstack[sp--]->inside(x,y,z, dx,dy,dz);
					else
						b = stack[sp--];
					if (!b) {
						// b=false => a|b = a
						if (bstack[sp]) {
							stack[sp]  = bstack[sp]->inside(x,y,z, dx,dy,dz);
							bstack[sp] = NULL;
						} // else already there no need to do anything
					} else {
						// b=true => a|b = true
						stack[sp]  = true;
						bstack[sp] = NULL;
					}
				}
				CHECK(cout << "  |  " << stack[sp] << endl);
			} else
			if (body == &GBody::tuniverse) {
				stack[++sp] = true;
				bstack[sp]  = NULL;
				CHECK(cout << "  @  " << stack[sp] << endl);
			} else {
				CHECK(int checked=cbody(engine, body)->isChecked());
				bstack[++sp] = cbody(engine, body);
				assert(sp<GZone::ZONE_STACK_SIZE);
				if (sp>=GZone::ZONE_STACK_SIZE) return false;
				CHECK(cout << " --> "<<body->name()
					   << " in["<<checked<<"]="
					   << "=" << stack[sp] << ":\t";
					for (int j=0; j<=sp; j++) cout << stack[j] << ' ';
					cout << endl);
			}
		}
		CHECK(cout << region()->name() << "=" << stack[0] << endl << endl);
		if (sp) return false;
		assert(sp==0);
		return stack[0];
	}
} // inside

/** inside*/
bool VZone::inside(GeometryEngine* engine,
		   const double  x, const double  y, const double  z,
		   const double dx, const double dy, const double dz,
		   const double  t) const
{
	if (std()) {
		int  i=0;
		// first plus terms (if any)
		while (i<size()) {
			const GBody* body = gexpr(i++);
			if (body == &GBody::tminus) break;
			if (!cbody(engine, body)->inside(x,y,z, dx,dy,dz, t)) return false;
		}
		// then negative terms (if any)
		while (i<size()) {
			const GBody* body = gexpr(i++);
			if (cbody(engine, body)->inside(x,y,z, dx,dy,dz, t)) return false;
		}
		return true;
	} else { // RPN
		bool stack[GZone::ZONE_STACK_SIZE], b;
		int  sp = -1;			// stack pointer
		// Perform a lazy(late) calculation only when it is needed
		CBody* bstack[GZone::ZONE_STACK_SIZE];	// NULL if already pushed to stack,
						// else needs evaluation of inside
		// Smart approach evaluating only the necessary terms
		// e.g.   True  or  ? => True
		//        False and ? => False
		for (int i=0; i<size(); i++) {
			const GBody* body = gexpr(i);
			// a+b = a && b
			if (body == &GBody::tplus) {
				assert(sp>0);
				if (!sp) return false;
				// check if any of the two is known and false
				if ((bstack[sp]==NULL   and !stack[sp]) or		// b
				    (bstack[sp-1]==NULL and !stack[sp-1])) {		// a
					// (a or b)=false => a+b = false
					sp--;
					stack[sp]  = false;
					bstack[sp] = NULL;
				} else {
					// retrieve b
					if (bstack[sp])
						b = bstack[sp--]->inside(x,y,z, dx,dy,dz, t);
					else
						b = stack[sp--];
					if (b) {
						// b = true => a+b = a
						if (bstack[sp]) {
							stack[sp]  = bstack[sp]->inside(x,y,z, dx,dy,dz, t);
							bstack[sp] = NULL;
						} // else already there no need to do anything
					} else {
						// b = false => a+b = false
						stack[sp]  = false;
						bstack[sp] = NULL;
					}
				}
			} else
			// a-b = a && !b
			if (body == &GBody::tminus) {
				assert(sp>0);
				if (!sp) return false;
				// check if any of the two is known and false
				if ((bstack[sp]==NULL   and  stack[sp]) or		//!b
				    (bstack[sp-1]==NULL and !stack[sp-1])) {		// a
					// (a or !b)=false => a-b = false
					sp--;
					stack[sp]  = false;
					bstack[sp] = NULL;
				} else {
					// retrieve b
					if (bstack[sp])
						b = bstack[sp--]->inside(x,y,z, dx,dy,dz, t);
					else
						b = stack[sp--];
					if (!b) {
						// b=false => a-b = a
						if (bstack[sp]) {
							stack[sp]  = bstack[sp]->inside(x,y,z, dx,dy,dz, t);
							bstack[sp] = NULL;
						} // else already there no need to do anything
					} else {
						// b=true => a-b = false
						stack[sp]  = false;
						bstack[sp] = NULL;
					}
				}
			} else
			// a | b
			if (body == &GBody::tunion) {
				assert(sp>0);
				if (!sp) return false;
				// check if any of the two is known and true
				if ((bstack[sp]==NULL   and  stack[sp]) or		// b
				    (bstack[sp-1]==NULL and  stack[sp-1])) {		// a
					// (a or b)=true => a|b = true
					sp--;
					stack[sp]  = true;
					bstack[sp] = NULL;
				} else {
					// retrieve b
					if (bstack[sp])
						b = bstack[sp--]->inside(x,y,z, dx,dy,dz, t);
					else
						b = stack[sp--];
					if (!b) {
						// b=false => a|b = a
						if (bstack[sp]) {
							stack[sp]  = bstack[sp]->inside(x,y,z, dx,dy,dz, t);
							bstack[sp] = NULL;
						} // else already there no need to do anything
					} else {
						// b=true => a|b = true
						stack[sp]  = true;
						bstack[sp] = NULL;
					}
				}
			} else
			if (body == &GBody::tuniverse) {
				stack[++sp] = true;
				bstack[sp]  = NULL;
			} else {
				bstack[++sp] = cbody(engine, body);
				assert(sp<GZone::ZONE_STACK_SIZE);
				if (sp>=GZone::ZONE_STACK_SIZE) return false;
			}
		}
		if (sp!=0) return false;
		return stack[0];
	}
} // inside

/** distance
 * @param x,y,z		location of ray
 * @param dx,dy,dz	normalized direction of ray
 * @param tmin		I/O minimum intersection point
 * @param tmax		maximum to scan
 * @return body that an intersection in found value with the minimum (*tmin)
 *
 * NOTE: Similar logic has the method Geometry::intersectRayUndefinedRegion
 */
CBody* VZone::distance(GeometryEngine* engine,
			   const double  x, const double  y, const double  z,
			   const double dx, const double dy, const double dz,
			   double* tmin, const double tmax) const
{
	while (true) {
		CBody* tbody = NULL;
		double t = tmax;

		for (int i=0; i<size(); i++) {
			const GBody* body = gexpr(i);
			if (!body->isBody()) continue;
			CBody* cb = cbody(engine, body);
			if (cb->distance(x,y,z, dx,dy,dz)) {
				if (InRangeOpen(*tmin, cb->tmin, t)) {
					t = cb->tmin * (1.0+SMALL);
					tbody = cb;
				} else
				if (InRangeOpen(*tmin, cb->tmax, t)) {
					t = cb->tmax * (1.0+SMALL);
					tbody = cb;
				}
			}
		}
		if (!tbody) {
			*tmin = tmax;
			return NULL;
		} else {
			// move it a bit
			// FIXME
			// It could be improved by getting rid of the SMALL3D
			// and find the next intersection (like the above loop
			// and check if it is inside or not
			t *= (1.0+SMALL);
			*tmin = t;
			if (!inside(engine, x,y,z, dx,dy,dz, t))
				return tbody;
		}
	}
} // distance

/** operator << */
std::ostream& operator << (std::ostream& s, const VZone& zone)
{
	s << zone.name() << ":" << zone.id();
	return s;
} /* operator << */
