/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Please look at the supplied documentation for the user's
 * license
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include <Python.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pyutils.h"
#include "gobject.h"
#include "viewerobject.h"

#include "xdraw/xdraw.h"

static const char *anchorStr[] = {
			"none", "C", "N",  "NE", "E",
			"SE",   "S", "SW", "W",  "NW" };

//static const char* objectTypeStr[]  = {
//		"default",	// Object_Default
//		"square",	// Point_Square
//		"cross",	// Point_Cross
//		"x",		// Point_X
//		"diamond",	// Point_Diamond
//		"circle",	// Point_Circle
//		"line",		// Arrow_Line
//		"head",		// Arrow_Head
//		"tail",		// Arrow_Tail
//		"headtail",	// Arrow_HeadTail
//		"simple",	// Ruler_Simple
//		"angle",	// Ruler_Angle
//		"xyz",		// Rotdefi_XYZ
//		"uvw",		// Rotdefi_UVW
//		"sun",		// Light_Sun
//		"omni",		// Light_Omni
//		"spot",		// Light_Spot
//		};

//static const char dottedPattern[2] = {1,1};
static const char dashedPattern[2] = {3,3};

// Selection color of items
dword GObject::nodeColor   = 0xFFB0B0;
dword GObject::selectColor = 0xFF00FF;

/* ================================ GObject ================================= */
/** Basic class for all graphics objects */
GObject::GObject(const std::string& aname, ObjectType ot)
	: _name(aname)
{
	type       = ot;
	anchor     = Anchor_none;
	color      = 0x000000;    // Color
	select     = false;       // Selected item
	selectNode = 0;           // None node is selected
	drawDw     = INFINITE;    // limit
	size       = 3;           // pixels
	lineWidth  = 1;           // default line width
	show       = false;       // default don't show anything

	xpos = ypos = 0;

	gcValues.function   = GXcopy;
	gcValues.line_width = lineWidth;
	gcValues.line_style = LineSolid;
} // GObject

/** setForeground */
inline void GObject::setForeground(ViewerObject* self) const
{
	self->xdraw.foreground(select? selectColor : color);
} // setForeground

/** drawPoint */
inline void GObject::drawPoint(ViewerObject* self, Drawable drawable,
				int ii, int jj, int s) const
{
	self->xdraw.rectangle(drawable, ii-s, jj-s, 2*s+1, 2*s+1);
	self->xdraw.point(drawable, ii, jj);
} // drawPoint

/** drawSelectedPoint */
inline void GObject::drawSelectedPoint(ViewerObject* self, Drawable drawable,
					int ii, int jj, int s) const
{
	self->xdraw.fillRectangle(drawable, ii-s, jj-s, 2*s+1, 2*s+1);
} // drawSelectedPoint

/** set parameters from dictionary
 * @param dict	dictionary with parameters to set
 * @param return true on error and set the python error
 * @return number of error parameters but not the names neither sets an exception
 */
int GObject::config(PyObject* dict)
{
	if (!PyDict_Check(dict)) {
		PyErr_SetString(PyExc_TypeError, "Invalid type, dictionary expected");
		return 1;
	}

	PyObject* key, *value;
#if PY_VERSION_HEX >= 0x02050300
	Py_ssize_t pos = 0;
#else
	int pos = 0;
#endif

	// Firstly set position if any
	value = PyDict_GetItemString(dict, "x");
	if (value) config("x", value);
	value = PyDict_GetItemString(dict, "y");
	if (value) config("y", value);
	value = PyDict_GetItemString(dict, "z");
	if (value) config("z", value);

	// loop over parameters
	int errcount = 0;
	while (PyDict_Next(dict, &pos, &key, &value))
		if (config(PyUnicode_AsUTF8(key), value) == NULL)
			errcount++;

	if (errcount) PyErr_Clear();

	return errcount;
} // config

/** set parameter
 * @param key	value to set
 * @param value	python object to set value from
 * @param return true on error and set the python error
 */
PyObject* GObject::config(const char* key, PyObject* value)
{
	if (!strcmp(key, "pos")) {
		if (value==NULL)
			return Py_Vector(P);
		else
			P = Py_GetVector(value);
	} else
	if (!strcmp(key, "move") && selectNode<=1) {
		if (value==NULL) {
			PyErr_SetString(PyExc_SyntaxError, "'move' do not return any value");
			return NULL;
		} else {
			Vector M = Py_GetVector(value);
			P += M;
		}
	} else
	if (!strcmp(key, "x")) {
		if (value==NULL)
			return PyFloat_FromDouble(P.x);
		else
			P.x = Py_GetFloat(value);
	} else
	if (!strcmp(key, "y")) {
		if (value==NULL)
			return PyFloat_FromDouble(P.y);
		else
			P.y = Py_GetFloat(value);
	} else
	if (!strcmp(key, "z")) {
		if (value==NULL)
			return PyFloat_FromDouble(P.z);
		else
			P.z = Py_GetFloat(value);
	} else
	if (!strcmp(key, "save"))
		save();
	else
	if (!strcmp(key, "restore"))
		restore();
	else
	if (!strcmp(key, "clearsave")) {
		selectNode = 0;
	} else
	if (!strcmp(key, "name")) {
		if (value==NULL)
			return PyUnicode_FromString(cname());
		else
			name(PyUnicode_AsUTF8(value));
	} else
	if (!strcmp(key, "move")) {
		Vector M = Py_GetVector(value);
		P += M;
	} else
	if (!strcmp(key, "id")) {
		if (value==NULL)
			return PyLong_FromLong(id());
		else {
			PyErr_SetString(PyExc_SyntaxError, "Cannot set the 'id'");
			return NULL;
		}
	} else
	if (!strcmp(key, "anchor")) {
		if (value==NULL)
			return PyUnicode_FromString(anchorStr[anchor]);
		else
		if (PyLong_Check(value))
			anchor = (AnchorType)PyLong_AsLong(value);
		else {
			const char* s = PyUnicode_AsUTF8(value);
			for (unsigned i=0; i<sizeof(anchorStr)/sizeof(char*); i++)
				if (!strcmp(s, anchorStr[i])) {
					anchor = (AnchorType)i;
					Py_RETURN_NONE;
				}
			PyErr_Format(PyExc_TypeError, "Invalid anchor \"%s\"",s);
			return NULL;
		}
	} else
	if (!strcmp(key, "color")) {
		if (value==NULL)
			return PyLong_FromLong((dword)color);
		else
		if (PyLong_Check(value))
			color = (dword)PyLong_AsLong(value);
		else
			color = (dword)strtol(PyUnicode_AsUTF8(value)+1, NULL, 16);
	} else
	if (!strcmp(key, "select")) {
		if (value==NULL)
			return PyLong_FromLong((dword)select);
		else {
			select = PyObject_IsTrue(value);
			if (!select) selectNode = 0;
		}
	} else
	if (!strcmp(key, "selectnode")) {
		if (value==NULL)
			return PyLong_FromLong(selectNode);
		else {
			if (select)
				selectNode = PyLong_AsLong(value);
			else
				selectNode = 0;
		}
	} else
	if (!strcmp(key, "selectcolor")) {
		if (value==NULL)
			return PyLong_FromLong((dword)selectColor);
		else {
			if (PyLong_Check(value))
				selectColor = (dword)PyLong_AsLong(value);
			else
				selectColor = (dword)strtol(PyUnicode_AsUTF8(value)+1, NULL, 16);
		}
	} else
	if (!strcmp(key, "nodecolor")) {
		if (value==NULL)
			return PyLong_FromLong((dword)nodeColor);
		else
		if (PyLong_Check(value))
			nodeColor = (dword)PyLong_AsLong(value);
		else
			nodeColor = (dword)strtol(PyUnicode_AsUTF8(value)+1, NULL, 16);
	} else
	if (!strcmp(key, "nodes")) {
		if (value==NULL)
			return PyLong_FromLong(nodes());
		else {
			PyErr_SetString(PyExc_SyntaxError, "Cannot set the number of nodes");
			return NULL;
		}
	} else
	if (!strcmp(key, "show"))
		if (value==NULL)
			return PyLong_FromLong((dword)show);
		else
			show = (bool)PyLong_AsLong(value);
	else
	if (!strcmp(key, "size")) {
		if (value==NULL)
			return PyLong_FromLong((long)size);
		else {
			size = Py_GetInt(value);
			if (size<=0) size = 5;
		}
	} else
	if (!strcmp(key, "linewidth")) {
		if (value==NULL)
			return PyLong_FromLong((long)lineWidth);
		else {
			lineWidth = Py_GetInt(value);
			if (lineWidth<=0) lineWidth = 0;
		}
	} else
	if (!strcmp(key, "drawDw")) {
		if (value==NULL)
			return PyFloat_FromDouble(drawDw);
		else
			drawDw = Py_GetFloat(value);
	} else
	if (!strcmp(key, "type") || !strcmp(key, "option")) {
		if (value==NULL)
			return PyLong_FromLong(type - (int)isA());
//			return PyUnicode_FromString(objectTypeStr[type]);
		else
		if (PyLong_Check(value))
			type = (ObjectType)(PyLong_AsLong(value) + (int)isA());
//		else {
//			const char* s = PyUnicode_AsUTF8(value);
//			for (unsigned i=0; i<sizeof(objectTypeStr)/sizeof(char*); i++)
//				if (!strcmp(s, objectTypeStr[i])) {
//					type = (ObjectType)i;
//					Py_RETURN_NONE;
//				}
//			PyErr_Format(PyExc_TypeError, "Invalid type \"%s\"",s);
//			return NULL;
//		}
	} else
	if (!strcmp(key, "class")) {
		if (value==NULL)
			return PyUnicode_FromString(typeStr());
		// No error is generated from this, since it is
		// set automatically during the initial config
	} else
	if (!strcmp(key, "dict")) {
		if (value==NULL) {
			PyErr_SetString(PyExc_SyntaxError, "'dict' do not return any value");
			return NULL;
		} else
			config(value);
	} else {
		PyErr_Format(PyExc_KeyError,"Object: Invalid type \"%s\"", key);
		return NULL;
	}

	if (PyErr_Occurred()) return NULL;
	Py_RETURN_NONE;
} // config

/** move object item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 */
void GObject::node(int, const Point& r)
{
	P = r;
} // node

/** transform coordinates to screen */
void GObject::transform(ViewerObject* self)
{
	self->kernel->view.xyz2uvw3D(P, &VP);
	VPc = VP;
	_visible = self->kernel->view.inside(VP.x,VP.y);
} // transform

/** @return bounding box of object */
BBox GObject::bbox() const
{
	BBox bb;
	bb.add(P);
	return bb;
} // bbox

/** @return bounding box of object */
BBox GObject::bboxView(ViewerObject* self)
{
	BBox bb;
	transform(self);
	bb.add(VP);
	return bb;
} // bboxView

/** draw */
void GObject::draw(ViewerObject* self, Drawable drawable)
{
	gcValues.line_style = LineSolid;
	gcValues.line_width = lineWidth;
	self->xdraw.changeGC(GCFunction|GCLineWidth|GCLineStyle, &gcValues);

	if (select && selectNode==NodeId_Pos and
	    _visible && self->kernel->view.inside(VP.x,VP.y)) {
		xpos = self->kernel->view.u2i(VP.x);
		ypos = self->kernel->view.v2j(VP.y);
		self->xdraw.foreground(nodeColor);
		drawPoint(self, drawable,
			self->kernel->view.u2i(VP.x),
			self->kernel->view.v2j(VP.y),
			size);
	}
	xpos = self->kernel->view.u2i(VPc.x);
	ypos = self->kernel->view.v2j(VPc.y);
	setForeground(self);
	if (!_name.empty() && anchor!=Anchor_none) drawText(self, drawable);
} // draw

/** drawText */
void GObject::drawText(ViewerObject* self, Drawable drawable)
{
	int i = xpos;
	int j = ypos;
	int l = (int)_name.length();
//	int VP.z = XTextWidth(font, name, l);

	switch (anchor) {
		case Anchor_N:
			j -= size+1;
			break;
		case Anchor_NE:
			j -= size+1;
			i += size+1;
			break;
		case Anchor_E:
			i += size+1;
			break;
		case Anchor_SE:
			j += size+1;
			i += size+1;
			break;
		case Anchor_S:
			j += size+1;
			break;
		case Anchor_SW:
			j += size+1;
			i -= size+1;
			break;
		case Anchor_W:
			i -= size+1;
			break;
		case Anchor_NW:
			j -= size+1;
			i -= size+1;
			break;
		default:	// just to keep compiler happy
			;
	}
	self->xdraw.string(drawable, i, j, cname(), l);
} // drawText

/** closest
 * Check if a point lies close to the object assuming also the size of the object
 * @param ii,jj	check point
 * @param d	check distance
 * @return NONE on failure,
 *          0 on object
 *          1 on main point (P)
 *          # index of item which is close to hit point ii,jj
 */
int GObject::closest(ViewerObject* self, const int ii, const int jj, const int d)
{
	xpos = self->kernel->view.u2i(VP.x);
	ypos = self->kernel->view.v2j(VP.y);

	if (close2Point(ii,jj, xpos,ypos, d))
		return NodeId_Pos;	// Main point
	else
		return NodeId_NONE;		// Nowhere
} // closest

/** return if pixel ii,jj is close to line described from viewer
 * points A and B
 * @param ii,jj	pixel on viewport
 * @param d	distance to check against to
 * @param A	first viewer point
 * @param B	second viewer point
 * @return	NodeId_NONE not close
 *		NodeId_0 close to line segment
 *		NodeId_1 close to A
 *		NodeId_2 close to B
 */
int GObject::close2Line(ViewerObject* self, const int ii, const int jj, const int d,
			const Point& A, const Point B) const
{
	int xa = self->kernel->view.u2i(A.x);
	int ya = self->kernel->view.v2j(A.y);

	if (close2Point(ii,jj, xa,ya, d)) return NodeId_1;	// Main point

	// Check other end
	int xb = self->kernel->view.u2i(B.x);
	int yb = self->kernel->view.v2j(B.y);
	if (close2Point(ii,jj, xb,yb, d)) return NodeId_2;	// End point

	double Dx = (double)(xb-xa);
	double Dy = (double)(yb-ya);
	double Dlen2 = Sq(Dx) + Sq(Dy);
	if (Dlen2 < (double)Sq(d)) return NodeId_NONE;

	double xx = (double)(ii-xa);
	double yy = (double)(jj-ya);

	// Check distance, calculate the cross product (to the square)
	// of the two vs length
	if (Sq(Dy*xx - Dx*yy) > Sq((double)(d)) * Dlen2)
		return NodeId_NONE;

	// Calculate the projection distance using the dot product
	if (InRange(0.0, Dx*xx + Dy*yy, Dlen2))
		return NodeId_0;

	return NodeId_NONE;
} // close2Line

/** enclosed */
bool GObject::enclosed(ViewerObject* self, const int left, const int top, const int right, const int bottom)
{
	xpos = self->kernel->view.u2i(VPc.x);
	ypos = self->kernel->view.v2j(VPc.y);

	int xl = Max(xpos-size, left);
	int yl = Max(ypos-size, top);
	int xh = Min(xpos+size, right);
	int yh = Min(ypos+size, bottom);

	return (xl<=xh && yl<=yh);
} // enclosed

/* ================================= GPoint ================================= */
/** draw */
void GPoint::draw(ViewerObject* self, Drawable drawable)
{
	int size2;
	XPoint	pts[5];

	GObject::draw(self, drawable);

	switch (type) {
		case Point_Dot:	// Draw a dot
			self->xdraw.point(drawable, xpos, ypos);
			break;

		case Point_Square:
			size2 = size*2 + 1;
			self->xdraw.rectangle(drawable, xpos-size, ypos-size, size2, size2);
			self->xdraw.point(drawable, xpos, ypos);
			break;

		case Point_X:
			self->xdraw.line(drawable, xpos-size, ypos-size, xpos+size+1, ypos+size+1);
			self->xdraw.line(drawable, xpos-size, ypos+size, xpos+size+1, ypos-size-1);
			break;

		case Point_Diamond:
			pts[0].x = pts[4].x = (short)(xpos);
			pts[0].y = pts[4].y = (short)(ypos - size);
			pts[1].x = (short)(xpos + size);
			pts[1].y = (short)(ypos);
			pts[2].x = (short)(xpos);
			pts[2].y = (short)(ypos + size);
			pts[3].x = (short)(xpos - size);
			pts[3].y = (short)(ypos);
			self->xdraw.lines(drawable, pts, 5, CoordModeOrigin);
			self->xdraw.point(drawable, xpos, ypos);
			break;

		case Point_Circle:
			size2 = size*2 + 1;
			self->xdraw.point(drawable, xpos, ypos);
			self->xdraw.arc(drawable, xpos-size, ypos-size, size2, size2, 0, 360*64);
			break;

		default:
			self->xdraw.line(drawable, xpos-size, ypos, xpos+size+1, ypos);
			self->xdraw.line(drawable, xpos, ypos-size, xpos, ypos+size+1);
	}
} // draw

/* ================================= GArrow ================================== */
/** set parameter
 * @param key	value to set
 * @param value	python object to set value from
 * @param return true on error and set the python error
 */
PyObject* GArrow::config(const char* key, PyObject* value)
{
	if (!strcmp(key, "pos")) {	// Override default behaviour of GObject
		switch (selectNode) {
			case 0:	// Whole arrow
				if (value==NULL)
					return Py_Vector(P);
				else
					P = Py_GetPoint(value);
				break;
			case 1: // Start point leaving end point untouched
				if (value==NULL)
					return Py_Vector(P);
				else {
					Point O = P;
					P = Py_GetPoint(value);
					D += O-P;
				}
				break;
			case 2:	// End point
				if (value==NULL)
					return Py_Vector(P+D);
				else {
					D = Py_GetVector(value);
					D -= P;
				}
				break;
		}
	} else
	if (!strcmp(key, "dx")) {
		if (value==NULL)
			return PyFloat_FromDouble(D.x);
		else
			D.x = Py_GetFloat(value);
	} else
	if (!strcmp(key, "dy")) {
		if (value==NULL)
			return PyFloat_FromDouble(D.y);
		else
			D.y = Py_GetFloat(value);
	} else
	if (!strcmp(key, "dz")) {
		if (value==NULL)
			return PyFloat_FromDouble(D.z);
		else
			D.z = Py_GetFloat(value);
	} else
	if (!strcmp(key, "dir")) {
		if (value==NULL)
			return Py_Vector(D);
		else
			D = Py_GetVector(value);
	} else
	if (!strcmp(key, "xe")) {
		if (value==NULL)
			return PyFloat_FromDouble(P.x+D.x);
		else
			D.x = Py_GetFloat(value) - P.x;
	} else
	if (!strcmp(key, "ye")) {
		if (value==NULL)
			return PyFloat_FromDouble(P.y+D.y);
		else
			D.y = Py_GetFloat(value) - P.y;
	} else
	if (!strcmp(key, "ze")) {
		if (value==NULL)
			return PyFloat_FromDouble(P.z+D.z);
		else
			D.z = Py_GetFloat(value) - P.z;
	} else
	if (!strcmp(key, "end")) {
		if (value==NULL)
			return Py_Vector(P+D);
		else {
			D = Py_GetVector(value);
			D -= P;
		}
	} else
	if (!strcmp(key, "type")) {
		PyObject* obj = GObject::config(key, value);
		switch (type) {
			case Arrow_HeadTail:
				drawHead = true;
				drawTail = true;
				break;
			case Arrow_Head:
				drawHead = true;
				drawTail = false;
				break;
			case Arrow_Tail:
				drawHead = false;
				drawTail = true;
				break;
			default:
				drawHead = false;
				drawTail = false;
				break;
		}
		return obj;
	} else
		return GObject::config(key, value);

	if (PyErr_Occurred()) return NULL;
	Py_RETURN_NONE;
} // config

/** @return node position */
Point GArrow::node(int item) const
{
	if (item==NodeId_Tail)
		return P + D;
	else
		return GObject::node(item);
} // node

/** move object item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 */
void GArrow::node(int item, const Point& r)
{
	switch (item) {
		case NodeId_Head:
			P = r;
			D = SP + SD - r;
			break;
		case NodeId_Tail:
			D  = r;
			D -= P;
			break;
		default:
			GObject::node(item,r);
	}
} // node

/** @return savedNode position */
Point GArrow::savedNode(int item) const
{
	if (item==2)
		return SP + SD;
	else
		return GObject::savedNode(item);
} // savedNode

/** transform */
void GArrow::transform(ViewerObject* self)
{
//	GObject::transform(self);
//	self->kernel->view.xyz2uvw3D(P+D, &Ve);
//	Vec = Ve;
//	_visible = self->kernel->view.clipLine(&VPc.x,&VPc.y, &Vec.x,&Vec.y);
	// FIXME WRONG VPc = VP
	_visible = self->kernel->view.clipLine3D(P, P+D, &VP, &Ve);
	VPc  = VP;
	Vec = Ve;
} // transform

/** rotate object around an axis
 * @param angle	to rotate
 * @param axis	to rotate
 */
void GArrow::rotate(const double angle, const Vector& axis)
{
	Matrix4	rot;
	rot.rotate(angle, axis);
	D = rot * SD;
} // rotate

/** save */
void GArrow::save()
{
	GObject::save();
	SD = D;
} // save

/** restore */
void GArrow::restore()
{
	GObject::restore();
	D = SD;
} // restore

/** @return bounding box of object */
BBox GArrow::bbox() const
{
	BBox bb;
	bb.add(P);
	bb.add(P+D);
	return bb;
} // bbox

/** @return bounding box of object */
BBox GArrow::bboxView(ViewerObject* self)
{
	BBox bb;
	transform(self);
	bb.add(VP);
	bb.add(Ve);
	return bb;
} // bboxView

/** draw */
void GArrow::draw(ViewerObject* self, Drawable drawable)
{
	GObject::draw(self, drawable);

	// Using the clipped relative-coordinates
	double x1d = self->kernel->view.u2id(VPc.x);
	double y1d = self->kernel->view.v2jd(VPc.y);
	double x2d = self->kernel->view.u2id(Vec.x);
	double y2d = self->kernel->view.v2jd(Vec.y);

	xpos = Round(x1d);
	ypos = Round(y1d);
	x2   = Round(x2d);
	y2   = Round(y2d);

	self->xdraw.line3D(drawable, self->kernel->view, VPc, Vec);

	// Draw mid-point
	self->xdraw.foreground(self->geometry->geometry->vertexColor);
	self->xdraw.point(drawable, (xpos+x2)/2, (ypos+y2)/2);
	setForeground(self);

	//self->xdraw.line(drawable, xpos, ypos, x2, y2);
	if (type==Arrow_Line) return;

	// Switch back to solid
	self->xdraw.setLineSolid();

	// Draw heads
	XPoint	pts[3];
	double ddx = x2d - x1d;
	double ddy = y2d - y1d;
	double length = hypot(ddx,ddy);
	if (length<0.00001) return;
	ddx /= length;
	ddy /= length;

	// Draw head  ----->
	if (self->kernel->view.inside(Ve.x,Ve.y)) {
		if (drawHead) {
			pts[0].x = (short)Round(x1d + ddx*(length-size) - ddy*size);
			pts[0].y = (short)Round(y1d + ddy*(length-size) + ddx*size);
			pts[1].x = (short)x2;
			pts[1].y = (short)y2;
			pts[2].x = (short)Round(x1d + ddx*(length-size) + ddy*size);
			pts[2].y = (short)Round(y1d + ddy*(length-size) - ddx*size);
			self->xdraw.lines(drawable, pts, 3, CoordModeOrigin);
		}
		if (select && selectNode==2 && self->kernel->view.inside(Ve.x,Ve.y)) {
			drawPoint(self, drawable,
				self->kernel->view.u2i(Ve.x),
				self->kernel->view.v2j(Ve.y),
				size);
		}
	}

	// Draw tail <-----
	if (drawTail && self->kernel->view.inside(VP.x,VP.y)) {
		pts[0].x = (short)Round(x1d + ddx*size - ddy*size);
		pts[0].y = (short)Round(y1d + ddy*size + ddx*size);
		pts[1].x = (short)xpos;
		pts[1].y = (short)ypos;
		pts[2].x = (short)Round(x1d + ddx*size + ddy*size);
		pts[2].y = (short)Round(y1d + ddy*size - ddx*size);
		self->xdraw.lines(drawable, pts, 3, CoordModeOrigin);
	}
} // draw

/** drawText */
void GArrow::drawText(ViewerObject* self, Drawable drawable)
{
	xpos = self->kernel->view.u2i(VPc.x);
	ypos = self->kernel->view.v2j(VPc.y);
	x2 = self->kernel->view.u2i(Vec.x);
	y2 = self->kernel->view.v2j(Vec.y);

	int i = (xpos+x2)/2;
	int j = (ypos+y2)/2;

	int l = (int)_name.length();
//	int VP.z = XTextWidth(font, name, l);

	switch (anchor) {
		case Anchor_N:
			j -= size+1;
			break;
		case Anchor_NE:
			j -= size+1;
			i += size+1;
			break;
		case Anchor_E:
			if (xpos>x2) {
				i = xpos + size+1;
				j = ypos;
			} else {
				i = x2 + size+1;
				j = y2;
			}
			break;
		case Anchor_SE:
			j += size+1;
			i += size+1;
			break;
		case Anchor_S:
			j += size+1;
			break;
		case Anchor_SW:
			j += size+1;
			i -= size+1;
			break;
		case Anchor_W:
			if (xpos<x2) {
				i = xpos - size+1;
				j = ypos;
			} else {
				i = x2 - size+1;
				j = y2;
			}
			break;
		case Anchor_NW:
			j -= size+1;
			i -= size+1;
			break;
		default:	// just to keep compiler happy
			;
	}
	self->xdraw.string(drawable, i, j, cname(), l);
} // drawText

/** return closest node for pixel (ii,jj) with distance < d
 * @param ii,jj	pixel on viewport
 * @param d	distance to check against to
 * @return	closest node
 */
int GArrow::closest(ViewerObject* self, const int ii, const int jj, const int d)
{
	return close2Line(self, ii,jj, size+d, VP, Ve);
} // closest

/** enclosed */
bool GArrow::enclosed(ViewerObject* self, const int left, const int top, const int right, const int bottom)
{
	xpos = self->kernel->view.u2i(VPc.x);
	ypos = self->kernel->view.v2j(VPc.y);
	x2 = self->kernel->view.u2i(Vec.x);
	y2 = self->kernel->view.v2j(Vec.y);

	ClipRegion clip(left,top, right,bottom);
	return clip.line(&xpos, &ypos, &x2, &y2);
} // enclosed

/* ================================= GSpline ================================= */
/** set parameter
 * @param key	value to set
 * @param value	python object to set value from
 * @param return true on error and set the python error
 */
PyObject* GSpline::config(const char* key, PyObject* value)
{
	if (!strcmp(key, "tension")) {
		if (value)
			spline.tension(PyFloat_AsDouble(value));
		else
			return PyFloat_FromDouble(spline.tension());
	} else
	if (!strcmp(key, "segments")) {
		if (value)
			nsegments = Range(1,(int)PyLong_AsLong(value),100);
		else
			return PyLong_FromLong(nsegments);
	} else
	if (!strcmp(key, "node")) {		// relative node position
		if (value==NULL)
			return PyLong_FromLong(spline.size());	// return size
		else
		if (PyLong_Check(value)) {			// (int) return node
			int item = PyLong_AsLong(value)-1;
			if (0<=item and item<(int)spline.size()) {
				Vector pos = spline[item];
				return Py_BuildValue("dddd",
						nodetime[item],
						pos.x,
						pos.y,
						pos.z);
			} else
			if (item<(int)nodetime.size())
				return PyFloat_FromDouble(nodetime[item]);
			else {
				PyErr_SetString(PyExc_IndexError, "invalid spline node index");
				return NULL;
			}
		} else
		if (Py_IsVector(value)) {
			double t = (nodetime.empty())? 0.0 : nodetime.back();
			spline.add(VectorSplineNode(Py_GetVector(value)));
			nodetime.push_back(t);
			scanDuplicates();
		} else
		if (PyTuple_Check(value)) {			// (tuple) set node,time
			// Accepting formats of:
			//  t			// time for last node
			//  Vector		// only position copy previous time
			//  x, y, z		// only position copy previous time
			//  t, Vector		// time and vector
			//  t, x, y, z		// time and vector components
			if (!InRange(1, (int)PyTuple_Size(value), 4)) {
				PyErr_SetString(PyExc_SyntaxError,
					"Wrong size of arguments, expected (t,vector), (t,x,y,z) or (t,) vector");
				return NULL;
			}

			double t = (nodetime.empty())? 0.0 : nodetime.back();
			Vector pos;

			if (PyTuple_Size(value)==1) {		// set last time
				if (Py_IsVector(PyTuple_GetItem(value, 0))) {
					// 1: vector
					pos = Py_GetVector(PyTuple_GetItem(value, 0));
					spline.add(VectorSplineNode(pos));
					scanDuplicates();
				} else
					// 1: t
					// only time ==> last node
					t = PyFloat_AsDouble(PyTuple_GetItem(value, 0));
				if (PyErr_Occurred()) return NULL;
				nodetime.push_back(t);
				Py_RETURN_NONE;
			} else
			if (PyTuple_Size(value)==2) {
				// 2: t,vector
				t = PyFloat_AsDouble(PyTuple_GetItem(value, 0));
				pos = Py_GetVector(PyTuple_GetItem(value, 1));
			} else {
				// 3: x,y,z or  4: t,x,y,z
				int i = 0;
				if (PyTuple_Size(value)==4)
					t = PyFloat_AsDouble(PyTuple_GetItem(value, i++));
				pos.x = PyFloat_AsDouble(PyTuple_GetItem(value, i++));
				pos.y = PyFloat_AsDouble(PyTuple_GetItem(value, i++));
				pos.z = PyFloat_AsDouble(PyTuple_GetItem(value, i++));
			}
			spline.add(VectorSplineNode(pos));
			nodetime.push_back(t);

			// check for duplicate point with previous one
			//int k=spline.size()-1;
			//if (k>0) {
			//	if (static_cast<Vector>(spline[k]).isClose(spline[k-1])) {
			//		spline[k].ignore = true;
			//		duplicates++;
			//	}
			//}
			//int oldduplicates = duplicates;
			scanDuplicates();
			//assert(duplicates==oldduplicates);
		} else {
			PyErr_SetString(PyExc_TypeError, "integer or tuple expected");
			return NULL;
		}
	} else
	if (!strcmp(key, "insert")) {
		if (PyTuple_Check(value) and PyTuple_Size(value)==2) {
			int item = PyLong_AsLong(PyTuple_GetItem(value,0))-1;
			Vector pos = Py_GetVector(PyTuple_GetItem(value,1));
			if (item<=spline.size()) {
				item = Max(0,item);
				spline.insert(item, VectorSplineNode(pos));
				nodetime.insert(nodetime.begin()+item, nodetime[item]);
			} else {
				spline.add(VectorSplineNode(Py_GetVector(value)));
				double t = (nodetime.empty())? 0.0 : nodetime.back();
				nodetime.push_back(t);
			}
			scanDuplicates();
		} else {
			PyErr_SetString(PyExc_TypeError, "tuple of size 2 expected (insert, position)");
			return NULL;
		}
	} else
	//if (!strcmp(key, "time")) {
	//	if (value and PyFloat_Check(value))
	//		nodetime.push_back(PyFloat_AsDouble(value));
	//	else
	//	if (value and PyLong_Check(value)) {
	//		int item = PyLong_AsLong(value);
	//		if (item<1 or item>(int)nodetime.size()) {
	//			PyErr_SetString(PyExc_IndexError, "invalid spline node index");
	//			return NULL;
	//		}
	//		return PyFloat_FromDouble(nodetime[item-1]);
	//	} else {
	//		PyErr_SetString(PyExc_TypeError, "floating point number expected");
	//		return NULL;
	//	}
	//} else
	if (!strcmp(key, "del")) {
		if (PyLong_Check(value)) {
			int item = PyLong_AsLong(value);
			if (item<1 or item>(int)spline.size()) {
				PyErr_SetString(PyExc_IndexError, "invalid spline node index");
				return NULL;
			}
			spline.del(item-1);
			nodetime.erase(nodetime.begin()+item-1);
			scanDuplicates();
		}
	} else
		return GObject::config(key, value);

	if (PyErr_Occurred()) return NULL;
	Py_RETURN_NONE;
} // config

/** scanDuplicates */
void GSpline::scanDuplicates()
{
	duplicates = 0;
	if (spline.size()>0) spline[0].ignore = false;
	for (int k=1; k<spline.size(); k++)
		if (static_cast<Vector>(spline[k]).isClose(spline[k-1])) {
			spline[k].ignore = true;
			duplicates++;
#ifdef STRONG
			std::cout << "DUP:" << k << " | total: " << duplicates << std::endl;
#endif
		} else
			spline[k].ignore = false;
} // scanDuplicates

/** save */
void GSpline::save()
{
	GObject::save();
	Spoint.resize(spline.size());
	for (int i=0; i<spline.size(); i++)
		Spoint[i] = spline[i];
} // save

/** restore */
void GSpline::restore()
{
	GObject::restore();
	for (int i=0; i<spline.size(); i++)
		spline[i] = Spoint[i];

	scanDuplicates();
} // restore

/** @return absolute savedNode position */
Point GSpline::savedNode(int item) const
{
	if (item<=0 || item>(int)Spoint.size()) return SP;
	assert(item-1 < (int)Spoint.size());
	return SP + Spoint[item-1];
} // savedNode

/** @return absolute node position */
Point GSpline::node(int item) const
{
	if (item<=0 || item>spline.size()) return P;
	assert(item-1 < spline.size());
	return P + spline[item-1];
} // node

/** set absolute node position */
void GSpline::node(int item, const Point& r)
{
	if (item<=0 || item>spline.size())
		GObject::node(0, r);
	else {
		assert(item-1 < spline.size());
		spline[item-1] = r-P;
		scanDuplicates();
	}
} // node

/** rotate */
//void GSpline::rotate(const double, const Vector& axis) {}
//{
//
//} // rotate

/** transform */
void GSpline::transform(ViewerObject* self)
{
	GObject::transform(self);

	int Nuniq = spline.size() - duplicates;
	if (spline.size()==0)
		return;
	else
	if (Nuniq==1) {
		// Single point draw a square
		ndrawsegs = 1;
		Vpoint.resize(1);
		self->kernel->view.xyz2uvw3D(P+spline[0], &Vpoint[0]);
	} else
	if (Nuniq==2 or (Nuniq==3 and type==Spline_Open)) {
		// Draw as lines
		ndrawsegs = 1;
		Vpoint.resize(spline.size());
		for (int i=0; i<spline.size(); i++)
			self->kernel->view.xyz2uvw3D(P+spline[i], &Vpoint[i]);
	} else {
		ndrawsegs = nsegments;
		if (Abs(spline.tension())<1e-10 or type > Spline_Closed)
			ndrawsegs = 1;

		int n = ndrawsegs*(Nuniq-1)+1;
		if (type==Spline_Closed or type==Spline_LinesClosed) {
			spline.closed = true;
			n += ndrawsegs;
		} else
			spline.closed = false;
		if (Vpoint.size() != (unsigned)n) Vpoint.resize(n);

		n = 0;
		double dt = 1.0/(double)ndrawsegs;
		for (int k=0; k<spline.size(); k++) {
			if (k+1<spline.size() and spline[k+1].ignore) continue;	// skip this node

			// node point
			assert(n < (int)Vpoint.size());
			self->kernel->view.xyz2uvw3D(P+spline[k], &Vpoint[n]);
			// Check only the node for visibility (not very clever)
			if (self->kernel->view.inside(Vpoint[k].x, Vpoint[n].y))
				_visible = true;
			n++;	// <<-- increase of the node point

			if (k+1<spline.size() or spline.closed) {
				double t = dt;
				for (int i=1; i<ndrawsegs; i++, t+=dt) {
					auto result = spline.evaluate(k, t);
					assert(n < (int)Vpoint.size());
					self->kernel->view.xyz2uvw3D(P+result, &Vpoint[n++]);
				}
			}
		}
		if (spline.closed) {
			assert(n < (int)Vpoint.size());
			Vpoint[n++] = Vpoint[0];
		}
	}
} // transform

/** @return bounding box of object */
BBox GSpline::bbox() const
{
	BBox bb = GObject::bbox();
	for (int i=0; i<spline.size(); i++)
		bb.add(P+spline[i]);
	return bb;
} // bbox

/** @return bounding box of object */
BBox GSpline::bboxView(ViewerObject* self)
{
	BBox bb = GObject::bboxView(self);
	for (unsigned i=0; i<Vpoint.size(); i++)
		bb.add(Vpoint[i]);
	return bb;
} // bboxView

/** draw */
void GSpline::draw(ViewerObject* self, Drawable drawable)
{
	GObject::draw(self, drawable);

	self->xdraw.setLineWidth(2);
	for (unsigned i=1; i<Vpoint.size(); i++)
		self->xdraw.line3D(drawable, self->kernel->view, Vpoint[i-1], Vpoint[i]);
	self->xdraw.setLineWidth(1);

	if (select) {
		int k = 0;
		int sel = selectNode-1;
		for (int i=0; i<spline.size(); i++) {
			if (spline[i].ignore) continue;
			int xn = self->kernel->view.u2i(Vpoint[k].x);
			int yn = self->kernel->view.v2j(Vpoint[k].y);
			if (i==sel)
				self->xdraw.fillRectangle(drawable, xn-size, yn-size, 2*size+1, 2*size+1);
			else
				self->xdraw.rectangle(drawable, xn-size, yn-size, 2*size+1, 2*size+1);

			char nodeid[10];
			sprintf(nodeid,"%d",i+1);
			self->xdraw.string(drawable, xn+size+6, yn, nodeid, strlen(nodeid));

			// if several nodes are very close together then we draw a circle
			// around them.
			// draw only the first if there are several sequential nodes that
			// are ignored because for the rest there is no correspondingly
			// transformed Vpoint!
			if (i+1 < spline.size() && spline[i+1].ignore) {
				xn = self->kernel->view.u2i(static_cast<Vector>(Vpoint[k]).x);
				yn = self->kernel->view.v2j(static_cast<Vector>(Vpoint[k]).y);
				self->xdraw.arc(drawable, xn-2*size, yn-2*size,
						size*4+1, size*4+1, 0, 360*64);
			}
			k += ndrawsegs;
		}
		//int x2 = self->kernel->view.u2i(Vpoint[0].x);
		//int y2 = self->kernel->view.v2j(Vpoint[0].y);
		//if (x2!=xpos or y2!=ypos) {
		//	self->xdraw.rectangle(drawable, xpos-size, ypos-size, 2*size+1, 2*size+1);
		//	self->xdraw.setDashes(0, dashedPattern, SIZE(dashedPattern));
		//	self->xdraw.line(drawable, xpos,ypos, x2,y2);
		//}
	}
} // draw

/** closest node to mouse location */
int GSpline::closest(ViewerObject* self, const int ii, const int jj, const int d)
{
	for (int i=0, k=0; i<spline.size(); i++) {
		if (spline[i].ignore) continue;
		int xn = self->kernel->view.u2i(Vpoint[k].x);
		int yn = self->kernel->view.v2j(Vpoint[k].y);
		if (Sq(ii-xn) + Sq(jj-yn) <= Sq(size+d))
			return i+1;
		k += ndrawsegs;
	}

	int knext = ndrawsegs;
	for (int k=1, i=1; k<(int)Vpoint.size(); k++) {
		if (close2Line(self, ii,jj,size+d, Vpoint[k-1], Vpoint[k]) != NodeId_NONE)
			return -i;
		if (k==knext) {
			do {
				i++;
			} while (i<spline.size() and spline[i].ignore);	// next node
			knext += ndrawsegs;
		}
	}

	return NodeId_NONE;
} // closest

/* ================================= GRuler ================================== */
/** set parameter
 * @param key	value to set
 * @param value	python object to set value from
 * @param return true on error and set the python error
 */
PyObject* GRuler::config(const char* key, PyObject* value)
{
	if (!strcmp(key, "xa")) {
		if (value==NULL)
			return PyFloat_FromDouble(P.x+Da.x);
		else
			Da.x = Py_GetFloat(value) - P.x;
	} else
	if (!strcmp(key, "ya")) {
		if (value==NULL)
			return PyFloat_FromDouble(P.y+Da.y);
		else
			Da.y = Py_GetFloat(value) - P.y;
	} else
	if (!strcmp(key, "za")) {
		if (value==NULL)
			return PyFloat_FromDouble(P.z+Da.z);
		else
			Da.z = Py_GetFloat(value) - P.z;
	} else
	if (!strcmp(key, "ang") || (!strcmp(key, "pos") && selectNode==3)) {
		// Override default behaviour of GObject
		if (value==NULL)
			return Py_Vector(P+Da);
		else {
			Da = Py_GetVector(value);
			Da -= P;
		}
	} else
	if (!strcmp(key, "dxa")) {
		if (value==NULL)
			return PyFloat_FromDouble(Da.x);
		else
			Da.x = Py_GetFloat(value);
	} else
	if (!strcmp(key, "dya")) {
		if (value==NULL)
			return PyFloat_FromDouble(Da.y);
		else
			Da.y = Py_GetFloat(value);
	} else
	if (!strcmp(key, "dza")) {
		if (value==NULL)
			return PyFloat_FromDouble(Da.z);
		else
			Da.z = Py_GetFloat(value);
	} else
	if (!strcmp(key, "dang")) {
		if (value==NULL)
			return Py_Vector(Da);
		else
			Da = Py_GetVector(value);
	} else
	if (!strcmp(key, "type")) {
		PyObject* obj = GObject::config(key, value);
		drawHead = true;
		drawTail = (type==Ruler_Simple);
		return obj;
	} else
		return GArrow::config(key, value);

	Py_RETURN_NONE;
} // config

/** @return node position */
Point GRuler::node(int item) const
{
	if (item==3)
		return P + Da;
	else
		return GArrow::node(item);
} // node

/** move object item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 */
void GRuler::node(int item, const Point& r)
{
	if (item==NodeId_Angle)
		Da  = r - P;
	else
		GArrow::node(item,r);
} // node

/** @return savedNode position */
Point GRuler::savedNode(int item) const
{
	if (item==NodeId_Angle)
		return SP + SDa;
	else
		return GArrow::savedNode(item);
} // savedNode

/** transform */
void GRuler::transform(ViewerObject* self)
{
	GArrow::transform(self);
	if (type == Ruler_Angle) {
		self->kernel->view.xyz2uvw3D(P+Da, &Va);
		Vac = Va;
		Vc2 = VP;
		_visible |= self->kernel->view.clipLine(&Vc2.x,&Vc2.y, &Vac.x,&Vac.y);
	}
} // transform

/** rotate object around an axis
 * @param angle	to rotate
 * @param axis	to rotate
 */
void GRuler::rotate(const double angle, const Vector& axis)
{
	Matrix4	rot;
	rot.rotate(angle, axis);
	D  = rot * SD;
	Da = rot * SDa;
} // rotate

/** save */
void GRuler::save()
{
	GArrow::save();
	SDa = Da;
} // save

/** restore */
void GRuler::restore()
{
	GArrow::restore();
	Da = SDa;
} // restore

/** @return bounding box of object */
BBox GRuler::bbox() const
{
	BBox bb;
	bb.add(P);
	bb.add(P+D);
	if (type == Ruler_Angle)
		bb.add(P+Da);
	return bb;
} // bbox

/** @return bounding box of object */
BBox GRuler::bboxView(ViewerObject* self)
{
	BBox bb;
	transform(self);
	bb.add(VP);
	bb.add(Ve);
	if (type == Ruler_Angle)
		bb.add(Va);
	return bb;
} // bboxView

/** draw */
void GRuler::draw(ViewerObject* self, Drawable drawable)
{
	gcValues.line_style = LineSolid;
	GArrow::draw(self, drawable);

	if (type==Ruler_Angle) {
		// Using the clipped relative-coordinates
		double x1d = self->kernel->view.u2id(Vc2.x);
		double y1d = self->kernel->view.v2jd(Vc2.y);
		xpos = Round(x1d);
		ypos = Round(y1d);

		double x3d = self->kernel->view.u2id(Vac.x);
		double y3d = self->kernel->view.v2jd(Vac.y);
		x3 = Round(x3d);
		y3 = Round(y3d);

		double x2d = self->kernel->view.u2id(Vec.x);
		double y2d = self->kernel->view.v2jd(Vec.y);

		// Draw end with solid line
		drawPoint(self, drawable, x3,y3, 3);

		// Calculate the angles to the reference axis and the end point
		double aa = atan2(y1d-y3d, x3d-x1d);
		double ae = atan2(y1d-y2d, x2d-x1d);
		double ang = ae-aa;
		//if (ang<0.0) ang += PI2;
		if (ang<0.0) {
			Swap(aa,ae);
			ang = -ang;
		}

		// Draw the angle arc
#define RADIUS	10
		self->xdraw.arc(drawable,
			xpos-RADIUS, ypos-RADIUS, 2*RADIUS, 2*RADIUS,
			Round(Deg(aa)) << 6, Round(Deg(ang)) << 6);

		// Draw selected point
		if (select && selectNode==3 && self->kernel->view.inside(Va.x,Va.y)) {
			drawPoint(self, drawable,
				self->kernel->view.u2i(Va.x),
				self->kernel->view.v2j(Va.y),
				3);
		}

		// And reference with dashed
		gcValues.line_width = 0;
		gcValues.line_style = LineOnOffDash;
		self->xdraw.changeGC(GCLineWidth|GCLineStyle, &gcValues);
		self->xdraw.setDashes(0, dashedPattern, SIZE(dashedPattern));
		self->xdraw.line(drawable, xpos, ypos, x3, y3);
	}
} // draw

/** drawText */
void GRuler::drawText(ViewerObject* self, Drawable drawable)
{
	char   txt[256];
	double Dlen = D.length();
	sprintf(txt, "%s: %.7g", cname(), Dlen);

	// Using the clipped relative-coordinates
	xpos = self->kernel->view.u2i(VPc.x);
	ypos = self->kernel->view.v2j(VPc.y);
	x2 = self->kernel->view.u2i(Vec.x);
	y2 = self->kernel->view.v2j(Vec.y);
	int i = (xpos+x2)/2;
	int j = (ypos+y2)/2;

	int l = (int)strlen(txt);
//	int VP.z = XTextWidth(font, txt, l);

	switch (anchor) {
		case Anchor_N:
			j -= size+1;
			break;
		case Anchor_NE:
			j -= size+1;
			i += size+1;
			break;
		case Anchor_E:
			if (xpos>x2) {
				i = xpos + size+1;
				j = ypos;
			} else {
				i = x2 + size+1;
				j = y2;
			}
			break;
		case Anchor_SE:
			j += size+1;
			i += size+1;
			break;
		case Anchor_S:
			j += size+1;
			break;
		case Anchor_SW:
			j += size+1;
			i -= size+1;
			break;
		case Anchor_W:
			if (xpos<x2) {
				i = xpos - size+1;
				j = ypos;
			} else {
				i = x2 - size+1;
				j = y2;
			}
			break;
		case Anchor_NW:
			j -= size+1;
			i -= size+1;
			break;
		default:	// just to keep compiler happy
			;
	}
	self->xdraw.string(drawable, i, j, txt, l);

	if (type==Ruler_Angle) {
		double Dalen = Da.length();
		double ang;
		if (Eq0(Dlen,SMALL) || Eq0(Dalen,SMALL))
			ang = 0.0;
		else
			ang = Deg(acos(Range(-1.0,(D*Da)/Dlen/Dalen,1.0)));
		sprintf(txt,"%.5g deg",ang);
		i = xpos;
		j = ypos;
		l = (int)strlen(txt);
		switch (anchor) {
			case Anchor_N:
				j -= size+1;
				break;
			case Anchor_NE:
				i += size+1;
				j -= size+1;
				break;
			case Anchor_E:
				i = xpos + size+1;
				//j = ypos;
				break;
			case Anchor_SE:
				i += size+1;
				j += size+1;
				break;
			case Anchor_S:
				j += size+1;
				break;
			case Anchor_SW:
				i -= size+1;
				j += size+1;
				break;
			case Anchor_W:
				i = xpos - size+1;
				//j = ypos;
				break;
			case Anchor_NW:
				i -= size+1;
				j -= size+1;
				break;
			default:	// just to keep compiler happy
				;
		}
		self->xdraw.string(drawable, i, j, txt, l);
	}
} // drawText

/** closest */
int GRuler::closest(ViewerObject* self, const int ii, const int jj, const int d)
{
	// First check angle point just in case is overlapped with the main one
	if (type==Ruler_Angle) {
		// Check angle point
		x3 = self->kernel->view.u2i(Va.x);
		y3 = self->kernel->view.v2j(Va.y);
		if (close2Point(ii,jj, x3,y3, d)) return NodeId_Angle;	// Angle point
	}

	// Check the arrow
	int close = GArrow::closest(self, ii, jj, d);
	if (close>=0) return close;

	// Check the arrow line
	if (type==Ruler_Angle and close2Line(self, ii,jj,size+d, Vc2, Vac) != NodeId_NONE)
		return NodeId_0;
	return NodeId_NONE;
} // closest

/** enclosed */
bool GRuler::enclosed(ViewerObject* self, const int left, const int top, const int right, const int bottom)
{
	if (GArrow::enclosed(self, left, top, right, bottom))
		return true;

	if (type==Ruler_Angle) {
		xpos = self->kernel->view.u2i(Vc2.x);
		ypos = self->kernel->view.v2j(Vc2.y);
		x2 = self->kernel->view.u2i(Vac.x);
		y2 = self->kernel->view.v2j(Vac.y);

		ClipRegion clip(left,top, right,bottom);
		return clip.line(&xpos, &ypos, &x2, &y2);
	} else
		return false;
} // enclosed

/* ================================= GRotdefi ================================== */
/** set parameter
 * @param key	value to set
 * @param value	python object to set value from
 * @param return true on error and set the python error
 */
PyObject* GRotdefi::config(const char* key, PyObject* value)
{
	if (!strcmp(key, "axissize")) {
		if (value==NULL)
			return PyLong_FromLong((long)axisSize);
		else {
			axisSize = Py_GetInt(value);
			if (axisSize<=0) axisSize = 20;
		}
	} else
	if (!strcmp(key, "axiswidth")) {
		if (value==NULL)
			return PyLong_FromLong((long)axisWidth);
		else {
			axisWidth = Py_GetInt(value);
			if (axisWidth<=0) axisWidth = 2;
		}
	} else
	if (!strcmp(key, "orient")) {
		if (value!=NULL) {
			if (!PyList_AsMatrix4(value, orient)) return NULL;
		} else
			return PyList_FromMatrix4(orient);
	} else
	if (!strcmp(key, "matrix")) {
		if (value!=NULL) {
			if (!PyList_AsMatrix4(value, matrix)) return NULL;
			D = matrix*P - P;
		} else
			return PyList_FromMatrix4(matrix);
	} else
		return GArrow::config(key, value);

	if (PyErr_Occurred()) return NULL;
	Py_RETURN_NONE;
} // config

/** draw */
void GRotdefi::draw(ViewerObject* self, Drawable drawable)
{
	D = matrix*P - P;
	GArrow::draw(self, drawable);

	if (axisWidth != lineWidth) {
		gcValues.line_width = axisWidth;
		self->xdraw.changeGC(GCLineWidth, &gcValues);
	}

	if (self->kernel->view.inside(Ve.x,Ve.y)) {
		Matrix4 A;
		A.multiply(self->kernel->view.matrix(), orient);
		self->xdraw.axes(drawable, xpos, ypos, axisSize, A);
	}

	if (self->kernel->view.inside(VP.x,VP.y)) {
		Matrix4 A;
		A.multiply(self->kernel->view.matrix(), matrix);
		self->xdraw.axes(drawable, x2, y2, axisSize, A, false);
	}
} // draw

/** closest */
int GRotdefi::closest(ViewerObject* self, const int ii, const int jj, const int d)
{
	int c = GArrow::closest(self, ii,jj, d);
	if (c==1) c = 0;
	return c;
} // closest

/** enclosed */
bool GRotdefi::enclosed(ViewerObject*  /*self*/, const int /*left*/, const int /*top*/, const int /*right*/, const int /*bottom*/)
{
/*
	int xpos = self->kernel->view.u2i(VP.x);
	int ypos = self->kernel->view.v2j(VP.y);

	int xl = Max(xpos-size, left);
	int yl = Max(ypos-size, top);
	int xh = Min(xpos+size, right);
	int yh = Min(ypos+size, bottom);

	return (xl<=xh && yl<=yh);
*/
	return false;
} // enclosed

/* ================================= GLight ================================= */
GLight::GLight(const char* aname, ObjectType o)
	: GArrow(aname, o)
{
	power    =  1.0;
	spec     =  0.0;
	relative = false;
	falloff  = Light::FallOff::Constant;
	shadow   = false;
	radius   = 0.0;
	nsamples = 1;
} // GLight

PyObject* GLight::config(const char* key, PyObject* value)
{
	if (!strcmp(key, "pos")) {	// Override default behaviour of GArrow
		if (selectNode<=1)
			return GObject::config(key,value);
		else {
			if (value==NULL)
				return Py_Vector(P+D);
			else {
				D = Py_GetVector(value);
				D -= P;
			}
		}
	} else
	if (!strcmp(key, "dx")) {
		if (value==NULL)
			return PyFloat_FromDouble(D.x);
		else
			D.x = Py_GetFloat(value);
	} else
	if (!strcmp(key, "dy")) {
		if (value==NULL)
			return PyFloat_FromDouble(D.y);
		else
			D.y = Py_GetFloat(value);
	} else
	if (!strcmp(key, "dz")) {
		if (value==NULL)
			return PyFloat_FromDouble(D.z);
		else
			D.z = Py_GetFloat(value);
	} else
	if (!strcmp(key, "relative")) {
		if (value==NULL)
			if (PyBool_FromLong(relative))
				return PyUnicode_FromString("on");
			else
				return PyUnicode_FromString("off");
		else
			relative = Py_GetBool(value);
	} else
	if (!strcmp(key, "power")) {
		if (value==NULL)
			return PyFloat_FromDouble(power);
		else
			power = Range(0.0, Py_GetFloat(value), 100000.0);
	} else
	if (!strcmp(key, "falloff")) {
		if (value==NULL)
			return PyLong_FromLong((int)falloff);
		else
		if (PyLong_Check(value)) {
			int f = PyLong_AsLong(value);
			if (f >= 3) {
				f -= 3;
				shadow  = false;
			} else
				shadow   = true;
			falloff = static_cast<Light::FallOff>(f);
		}
	} else
	if (!strcmp(key, "specular")) {
		if (value==NULL)
			return PyFloat_FromDouble(spec);
		else
			spec = Range(0.0, Py_GetFloat(value), 100000.0);
	} else
	if (!strcmp(key, "shadow")) {
		if (value==NULL)
			if (PyBool_FromLong(shadow))
				return PyUnicode_FromString("on");
			else
				return PyUnicode_FromString("off");
		else
			shadow = Py_GetBool(value);
	} else
	if (!strcmp(key, "nsamples")) {
		if (value==NULL)
			return PyLong_FromLong(nsamples);
		else
			nsamples = Range(0, Py_GetInt(value), 100);
	} else
	if (!strcmp(key, "radius")) {
		if (value==NULL)
			return PyFloat_FromDouble(radius);
		else
			radius = Range(0.0, Py_GetFloat(value), 100000.0);
	} else
		return GArrow::config(key, value);

	Py_RETURN_NONE;
} // config

/** move object item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 */
void GLight::node(int item, const Point& r)
{
	if (item <= NodeId_Pos)
		P = r;
	else
		GArrow::node(item,r);
} // node

/** transform *
void GLight::transform(ViewerObject* self)
{
	if (relative) {
		VP.x  = x;
		VP.y  = y;
		VP.z  = z;
		du = D.x;
		dv = D.y;
		dw = D.z;
	} else {
		GObject::transform(self);
		self->kernel->view.dxyz2duvw3D(D.x, D.y, D.z, &du, &dv, &dw);
	}

	_visible = self->kernel->view.inside(VP.x,VP.y);
} * transform */

/** draw */
void GLight::draw(ViewerObject* self, Drawable drawable)
{
	int size2;

	if (type != Light_Omni)
		GArrow::draw(self, drawable);
	else {
		GObject::draw(self, drawable);
		// unfortunatelly the point is on different plane from the maximum circle
		//if (select && selectNode==2 && self->kernel->view.inside(Ve.x,Ve.y)) {
		//	drawPoint(self, drawable,
		//		self->kernel->view.u2i(Ve.x),
		//		self->kernel->view.v2j(Ve.y),
		//		3);
		//}
	}

	size2 = size*2 + 1;
	self->xdraw.arc(drawable, xpos-size, ypos-size, size2, size2, 0, 360*64);
	gcValues.line_style = LineOnOffDash;
	self->xdraw.changeGC(GCLineStyle, &gcValues);
	self->xdraw.setDashes(0, dashedPattern, SIZE(dashedPattern));
	self->xdraw.arc(drawable, xpos-size-2, ypos-size-2, size2+4, size2+4, 0, 360*64);

	if (type==Light_Omni && select) {
		double R = D.length() * self->kernel->view.Sx();
		if (R<1e4) {
			int r = (int)R;
			gcValues.line_style = LineSolid;
			self->xdraw.changeGC(GCLineStyle, &gcValues);
			self->xdraw.arc(drawable, xpos-r, ypos-r, 2*r, 2*r, 0, 360*64);
		}
	}
} // draw

/** drawText */
void GLight::drawText(ViewerObject* self, Drawable drawable)
{
	if (type != Light_Omni)
		GArrow::drawText(self, drawable);
	else
		GObject::drawText(self, drawable);
} // drawText

/** closest */
int GLight::closest(ViewerObject* self, const int ii, const int jj, const int d)
{
	int close = GArrow::closest(self,ii,jj,d);
	if (type == Light_Omni) {
		double R = D.length() * self->kernel->view.Sx();
		if (R<1e9 && Abs(sqrt(Sq(ii-xpos) + Sq(jj-ypos)) - (int)R) < d)
			return NodeId_Radius;
	}
	return close;
} // closest

/** complete light structure */
void GLight::toLight(Light& L) const
{
	switch (type) {
		case Light_Sun:
			L.type = Light::Type::Sun;
			break;
		case Light_Omni:
			L.type = Light::Type::Omni;
			break;
		case Light_Spot:
			L.type = Light::Type::Spot;
			break;
		default:
			L.type = Light::Type::Sun;
	}
	L.pos      = P;
	L.dir      = D;
	L.distance = L.dir.normalize();
	L.relative = relative;
	L.power    = power;
	L.falloff  = falloff;
	L.spec     = spec;
	L.shadow   = shadow;
	if (radius<=0.0) {
		L.radius   = 0.0;
		L.nsamples = 1;
	} else {
		L.radius   = radius;
		L.nsamples = nsamples;
	}
} // GLight

/* ================================= GBeam ================================= */
GBeam::GBeam(const char* aname, ObjectType o)
	: GArrow(aname, o)
{
	energy      = 1.0;
	scale       = 1.0;
	divergence2 = 0.0;
	Rin = Rout  = 0.0;
} // GBeam

/** config */
PyObject* GBeam::config(const char* key, PyObject* value)
{
	if (!strcmp(key, "energy")) {
		if (value==NULL)
			return PyFloat_FromDouble((double)energy);
		else {
			energy = Abs(Py_GetFloat(value));
			if (energy<=0.0) energy=1.0;
		}
	} else
	if (!strcmp(key, "scale")) {
		if (value==NULL)
			return PyFloat_FromDouble((double)scale);
		else {
			scale = Abs(Py_GetFloat(value));
			if (scale<=0.0) scale=1.0;
		}
	} else
	if (!strcmp(key, "divergence")) {
		if (value==NULL)
			return PyFloat_FromDouble(2.0*divergence2);
		else
			divergence2 = Abs(Py_GetFloat(value))/2.0;
	} else
	if (!strcmp(key, "dx")) {
		if (value==NULL)
			return PyFloat_FromDouble(D.x/(energy*scale));
		else
			D.x = Py_GetFloat(value)*energy*scale;
	} else
	if (!strcmp(key, "dy")) {
		if (value==NULL)
			return PyFloat_FromDouble(D.y/(energy*scale));
		else
			D.y = Py_GetFloat(value)*energy*scale;
	} else
	if (!strcmp(key, "dz")) {
		if (value==NULL)
			return PyFloat_FromDouble(D.z/(energy*scale));
		else
			D.z = Py_GetFloat(value)*energy*scale;
	} else
	if (!strcmp(key, "matrix")) {
		if (value==NULL) {
			PyErr_SetString(PyExc_ValueError,"GBeam: cannot return beam matrix");
			return NULL;
		} else {
			Matrix4 matrix;
			PyList_AsMatrix4(value, matrix);
			P *= matrix;
			D *= matrix;
		}
	} else
		return GArrow::config(key, value);

	Py_RETURN_NONE;
} // config

/** draw */
void GBeam::draw(ViewerObject* self, Drawable drawable)
{
	GArrow::draw(self, drawable);

	//if (type==Beam_Divergence && divergence2>1E-6) {
	if (divergence2>1E-6) {
		int dx = x2-xpos;
		int dy = y2-ypos;	// up-side down screen
		int length = (int)(sqrt(dx*dx + dy*dy)*0.8);
		double angle = atan2(dy, dx);

		if (divergence2<PI2) {
			XPoint	pts[3];
			double s, c;
			sincos(angle-divergence2, &s, &c);
			pts[0].x = xpos+(int)(c*length);
			pts[0].y = ypos+(int)(s*length);

			pts[1].x = xpos;
			pts[1].y = ypos;

			sincos(angle+divergence2, &s, &c);
			pts[2].x = xpos+(int)(c*length);
			pts[2].y = ypos+(int)(s*length);

			self->xdraw.lines(drawable,
					pts, SIZE(pts), CoordModeOrigin);
		}
		int angle1 = (int)(Deg(-angle-divergence2)*64.0);
		int angle2 = (int)(Deg(divergence2)*2.0*64.0);

		self->xdraw.arc(drawable,
				xpos-length, ypos-length, 2*length, 2*length,
				angle1, angle2);
	}
	GObject::drawText(self, drawable);
} // draw

// ================================== GCamera ==================================
/** initialize vectors */
void GCamera::init()
{
	Z = at;
	if (Z.normalize()<1e-10) Z = Vector::Zo;

	//Vector Yold = Y;
	Y = up;
	if (Y.normalize()<1e-10) Y = Vector::Yo;
	if (!Eq(Abs(Z*Y),1.0,1e-5)) {
		Z.gramSchmidt(Y);
		Y.normalize();
		X = Y ^ Z;
	} else {
		// In case Z // (Y=up) then use previous information if any
		Y = Y.orthogonal();
		Z.gramSchmidt(Y);
		Y.normalize();
		X = Y ^ Z;
	}

	w2 = focal*tan(fov/2);
	h2 = w2/aspect;
} // init

/** set parameters
 * @param key	value to set
 * @param value	python object to set value from
 * @param return true on error and set the python error
 */
PyObject* GCamera::config(const char* key, PyObject* value)
{
	if (!strcmp(key, "atx")) {
		if (value==NULL)
			return PyFloat_FromDouble(P.x+at.x);
		else
			at.x = Py_GetFloat(value) - P.x;
	} else
	if (!strcmp(key, "aty")) {
		if (value==NULL)
			return PyFloat_FromDouble(P.y+at.y);
		else
			at.y = Py_GetFloat(value) - P.y;
	} else
	if (!strcmp(key, "atz")) {
		if (value==NULL)
			return PyFloat_FromDouble(P.z+at.z);
		else
			at.z = Py_GetFloat(value) - P.z;
	} else
	if (!strcmp(key, "upx")) {
		if (value==NULL)
			return PyFloat_FromDouble(up.x);
		else
			up.x = Py_GetFloat(value);
	} else
	if (!strcmp(key, "upy")) {
		if (value==NULL)
			return PyFloat_FromDouble(up.y);
		else
			up.y = Py_GetFloat(value);
	} else
	if (!strcmp(key, "upz")) {
		if (value==NULL)
			return PyFloat_FromDouble(up.z);
		else
			up.z = Py_GetFloat(value);
	} else
	if (!strcmp(key, "focal")) {
		if (value==NULL)
			return PyFloat_FromDouble(focal);
		else
			focal = Max(0.0,Py_GetFloat(value));
	} else
	if (!strcmp(key, "fov")) {
		if (value==NULL)
			return PyFloat_FromDouble(Deg(fov));
		else {
			double v = Py_GetFloat(value);
			if (v<=0.0) v = 45;
			fov = Range(0.001, Rad(v), 0.999*PI);
		}
	} else
	if (!strcmp(key, "aspect")) {
		if (value==NULL)
			return PyFloat_FromDouble(aspect);
		else {
			double v = Py_GetFloat(value);
			if (v<=0.0) v = 4./3.;
			aspect = Range(0.1, v, 10.0);
		}
	//} else
	//if (!strcmp(key, "type")) {
	//	PyObject* obj = GObject::config(key, value);
	//	drawTail = (type==Ruler_Simple);
	//	return obj;
	} else {
		PyObject* rc = GObject::config(key, value);
		init();
		return rc;
	}
	init();
	Py_RETURN_NONE;
} // config

/** save */
void GCamera::save()
{
	GObject::save();
	Sat    = at;
	Sup    = up;
	Sfocal = focal;
	SY     = Y;
	SZ     = Z;
} // save

/** restore */
void GCamera::restore()
{
	GObject::restore();
	at    = Sat;
	up    = Sup;
	focal = Sfocal;
	Y     = SY;
	Z     = SZ;
} // restore

/** @return node position */
Point GCamera::node(int item) const
{
	switch (item) {
		case NodeId_At:
			return P + at;
		case NodeId_Focal:
			return P + Z*focal;
		case NodeId_Up:
			return P + 2*h2*up;
		default:
			return GObject::node(item);
	}
} // node

/** @return savedNode position */
Point GCamera::savedNode(int item) const
{
	switch (item) {
		case NodeId_At:
			return SP + Sat;
		case NodeId_Focal:
			return P + SZ*Sfocal;
		case NodeId_Up:
			return SP + 2*h2*Sup;
		default:
			return GObject::savedNode(item);
	}
} // savedNode

/** move object item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 */
void GCamera::node(int item, const Point& r)
{
	switch (item) {
		case NodeId_Pos:
			P = r;
			at = (SP+Sat) - r;
			break;
		case NodeId_At:
			at = r - SP;
			break;
		case NodeId_Focal:
			focal = Max(0.0, (r-SP) * Z);
			break;
		case NodeId_Up:
			up = (r - SP);
			up.normalize();
			break;
		default:
			GObject::node(item,r);
	}
	init();
} // node

/** transform */
void GCamera::transform(ViewerObject* self)
{
	GObject::transform(self);

	self->kernel->view.xyz2uvw3D(P+at, &Vat);
	Vatc = Vat;
	_visible = self->kernel->view.clipLine(&VPc.x,&VPc.y, &Vatc.x,&Vatc.y);
	//self->kernel->view.xyz2uvw3D(P+up, &Vup);

	// viewer drawings points
	Point Pz = P + focal*Z;
	self->kernel->view.xyz2uvw3D(Pz + w2*X + h2*Y, &(Vpoint[0]));
	self->kernel->view.xyz2uvw3D(Pz - w2*X + h2*Y, &(Vpoint[1]));
	self->kernel->view.xyz2uvw3D(Pz - w2*X - h2*Y, &(Vpoint[2]));
	self->kernel->view.xyz2uvw3D(Pz + w2*X - h2*Y, &(Vpoint[3]));
	self->kernel->view.xyz2uvw3D(Pz + 1.5*h2*Y, &(Vpoint[4]));

} // transform

/** draw */
void GCamera::draw(ViewerObject* self, Drawable drawable)
{
	GObject::draw(self, drawable);

	// Draw camera
	for (int i=0; i<4; i++) {
		self->xdraw.line3D(drawable, self->kernel->view, VPc, Vpoint[i]);
		self->xdraw.line3D(drawable, self->kernel->view, Vpoint[i], Vpoint[(i+1)%4]);
	}
	self->xdraw.line3D(drawable, self->kernel->view, Vpoint[0], Vpoint[4]);
	self->xdraw.line3D(drawable, self->kernel->view, Vpoint[1], Vpoint[4]);

	if (select) {
		// the master point
		drawPoint(self, drawable,
				self->kernel->view.u2i(VPc.x),
				self->kernel->view.v2j(VPc.y),
				3);
		// 'at' marker if selected
		drawPoint(self, drawable,
				self->kernel->view.u2i(Vatc.x),
				self->kernel->view.v2j(Vatc.y),
				3);
	}

	// And reference with dashed
	gcValues.line_width = 0;
	gcValues.line_style = LineOnOffDash;
	self->xdraw.changeGC(GCLineWidth|GCLineStyle, &gcValues);
	self->xdraw.setDashes(0, dashedPattern, SIZE(dashedPattern));
	self->xdraw.line3D(drawable, self->kernel->view, VPc, Vatc);

	// Draw up vector
	//std::cout << "selectcolor node=" << selectNode << std::endl;
	if (select) { // and _visible) {
		Point U;
		self->kernel->view.xyz2uvw3D(P+2*h2*up, &U);
		self->xdraw.line3D(drawable, self->kernel->view, VP, U);
		drawPoint(self, drawable,
				self->kernel->view.u2i(U.x),
				self->kernel->view.v2j(U.y),
				3);
		//switch (selectNode) {
		//}
	}
} // draw

/** closest node to mouse location */
int GCamera::closest(ViewerObject* self, const int ii, const int jj, const int d)
{
	int close = close2Line(self, ii,jj,size+d, VP, Vatc);
	if (close>NodeId_0) return close;

	Point F;			// focal point on viewport
	self->kernel->view.xyz2uvw3D(P+focal*Z, &F);
	int xf = self->kernel->view.u2i(F.x);
	int yf = self->kernel->view.v2j(F.y);

	if (close2Point(ii,jj, xf,yf, d)) return NodeId_Focal;	// focal length

	if (select and _visible) {
		Point U;
		self->kernel->view.xyz2uvw3D(P+2*h2*up, &U);
		xf = self->kernel->view.u2i(U.x);
		yf = self->kernel->view.v2j(U.y);
		if (close2Point(ii,jj, xf,yf, d)) return NodeId_Up;	// up vector
	}

	// Check any of the drawing lines of the screen
	for (int k=0; k<4; k++) {
		if (close2Line(self, ii,jj,size+d, Vpoint[k], Vpoint[(k+1)%4]) != NodeId_NONE)
			return NodeId_Focal;
		if (close2Line(self, ii,jj,size+d, VP, Vpoint[k]) != NodeId_NONE)
			return NodeId_0;
	}
	// hat?
	if (close2Line(self, ii,jj,size+d, Vpoint[0], Vpoint[4]) != NodeId_NONE or
	    close2Line(self, ii,jj,size+d, Vpoint[1], Vpoint[4]) != NodeId_NONE)
			return NodeId_Focal;

	return close;
} // closest
