/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	16-Jul-2010
 */

#ifndef __GRAPH_PAINTER_H
#define __GRAPH_PAINTER_H

#include "bfont.h"
#include "color.h"
#include "clipregion.h"
#include "math/bmath.h"
#include "geometry/geo.h"

// ================================== IPoint ===================================
// FIXME remove from here => make more general if needed
/** integer 2D point class */
class IPoint {
public:
	int	x, y;

public:
	IPoint() : x(0), y(0) {}
	IPoint(int i, int j) : x(i), y(j) {}
	void	set(int i, int j) {
			x = i;
			y = j;
		}
}; // class IPoint

// ================================== Painter ==================================
/** Painter base class
 * basic painter class, provide means of clipping lines
 * Holds the data of an Image ready to be used as XImage
 * Default color depth is 24, therefore bits per pixel 32
 */
class Painter {
public:
	enum class Fill {		// Fill types/patterns
		FLOOD,
		DOTS,
		HASH,
		HASHR,
		X,
		X2
	};

	enum class Fit {		// Fit types
		SCALE,			// simple fast scaling
		//SCALE_INTERPOLATE	// scale and interpolate if needed
		//CROP			// Center and crop to the biggest dimension
	};

protected:
	// use signed types to avoid problems when clipping negative points
	int		_width;		/** width in pixels		*/
	int		_height;	/** height in pixels		*/
	int		_dataSize;	/** width*height		*/
	int		_maxSize;	/** allocated memory size dwords*/
	dword*		_data;		/** pointer to allocated memory	*/
	dword		_background;	/** background (gray) color	*/
	ClipRegion	_clip;		/** clip region			*/

public:
	Painter();
	~Painter();

	bool	init(const int w, const int h);

	// Data WARNING needed to deallocate from X11
	void	dataNull()		{ _data = NULL; }
	dword*	data()			{ return _data; }

	int	width()		const	{ return _width; }
	int	height()	const	{ return _height; }

	/** @param byte	background byte (only gray scale are allowed) */
	void	background(dword b)	{ _background = b; }
	dword	background()	const	{ return _background; }

	// Clipping
	void	resetClip()		{ _clip.set(_width-1, _height-1); }
	void	clip(int l, int t, int r, int b)
			{ _clip.set(l,t,r,b, _width-1, _height-1); }
const	ClipRegion& clip()	const	{ return _clip; }
	bool	insideClip(int x, int y) const	{ return _clip.inside(x,y); }

	// Drawing
	void	clear();
	// Warning UNSAFE operations
	void	pixel(const int x, const int y, dword color)
			{ _data[y*_width+x] = color; }
	void	pixel(const int x, const int y, dword color, double intensity);
	dword	pixel(const int x, const int y) const
			{ return _data[y*_width+x]; }
	dword&	pixel(const int x, const int y)
			{ return _data[y*_width+x]; }
	dword	operator()(const int x, const int y)	const	{ return pixel(x,y); }
	dword*	pixelPtr(const int x, const int y)
			{ return _data + y*_width+x; }

	// Clipped operations SAFE
	void	setPixel(const int x, const int y, dword color)
			{ if (_clip.inside(x,y)) _data[y*_width+x] = color; }
	dword	getPixel(const int x, const int y) const
			{ return _clip.inside(x,y)? _data[y*_width+x] : background(); }

	// Move image
	void	move(const int dx, const int dy);

	/** draw a line
	 * @param x1,y1	starting point of line
	 * @param x2,y2	end point of line
	 * @param color	color of line
	 * @return true if line is inside clip area
	 */
	bool	line(int x1, int y1, int x2, int y2, const dword color) {
			if (_clip.line(&x1, &y1, &x2, &y2)) {
				unclippedLine(x1,y1,x2,y2,color);
				return true;
			}
			return false;
		}
	bool	lineThick(int x1, int y1, int x2, int y2, const int w, const dword color) {
			if (_clip.line(&x1, &y1, &x2, &y2)) {
				unclippedThickLine(x1,y1, x2,y2, w, color);
				return true;
			}
			return false;
		}
	bool	lineAntialias(int x1, int y1, int x2, int y2, const dword color, bool alpha=false) {
			if (_clip.line(&x1, &y1, &x2, &y2)) {
				unclippedLineAntialias(x1,y1, x2,y2, color, alpha);
				//unclippedLine(x1,y1, x2,y2, color);
				//unclippedLineBeautyPass(x1,y1, x2,y2, color);
				return true;
			}
			return false;
		}
	bool	lineBeautyPass(int x1, int y1, int x2, int y2, const dword color) {
			if (_clip.line(&x1, &y1, &x2, &y2)) {
				unclippedLineBeautyPass(x1,y1, x2,y2, color);
				return true;
			}
			return false;
		}
	bool	drawBitmap(int, int, int, int, const dword *);

	bool	rectangle(int, int, int, int, const dword);
	void	fillRect(int, int, int, int, const dword);
	void	levelShiftRect(int, int, int, int, int);
	void	fill(int, int, const dword, const dword = 0, const Fill filltype=Fill::FLOOD);

	void	copy(Painter& src);
	void	fit(Painter& src, Fit fittype=Fit::SCALE);

	int	measure(BFont& font, const char *str) { return font.measure(str); }
	int	measuref(BFont&, const char *, ...);
	int	printf(BFont&, int, int, const dword, const char *, ...);
	int	vprintf(BFont&, int, int, const dword, const char *, ...);
	int	printfShadow(BFont&, int, int, int, const dword, const dword, const char *, ...);
	int	printfOutline(BFont&, int, int, const dword, const dword, const char *, ...);
	void	drawchar(BFont& font, int x, int y, const dword color, const char ch)
			{ font.draw(*this, x, y, color, ch); }
	void	vdrawchar(BFont& font, int x, int y, const dword color, const char ch)
			{ font.vdraw(*this, x, y, color, ch); }

	// second pass to fix missed anti-aliased pixels
	void	correctAntialias();
	// run anti-aliasing + any other post-processing filter that might be wanted
	void	gaussianBlur();

	size_t	memory() const { return sizeof(Painter) + _maxSize*sizeof(dword); }

protected:
	void	unclippedLine(int x1, int y1, int x2, int y2, const dword color);
	void	unclippedThickLine(int x1, int y1, int x2, int y2, const int w, const dword color);
	void	unclippedLineAntialias(int x1, int y1, int x2, int y2, dword lineColor, bool alpha);
	void	unclippedLineBeautyPass(int x1, int y1, int x2, int y2, dword lineColor);

	// convolute pixel with a specific kernel (e.g. Gaussian blur) and return the new color
	// NOTE: kernels MUST be symmetric!
	template<int N, int M>
		dword convolutePixel(int x, int y, const float (&pKernel)[N][M]) {
			Color newColor;
			assert(N == M);
			for (int i=0; i<N; i++) {
				for (int j=0; j<N; j++) {
					int curX = x - N/2 + i;
					int curY = y - N/2 + j;
					// for edges we introduce a slight error here due to the
					// wrap-around, but it's hardly noticeable and avoids
					// branching which might slow down the process
					int dataPtr = Clamp<int>(curY*_width + curX, 0, _width*_height - 1);
					Color curColor(_data[dataPtr]);
					newColor += curColor * pKernel[i][j];
				}
			}
			return newColor.color32();
		}
}; // Painter

#endif
