/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	31-Mar-2010
 */

#ifndef __MATH_BMATH_H
#define __MATH_BMATH_H

#include <math.h>

#include <limits>
#include <iostream>

#include "os.h"

#define MAX_REAL	1e308

const double PI         = 3.141592653589793238462643383279;
const double PI2        = 6.283185307179586476925286766559;
const double PIover2    = 1.570796326794896619231321691640;
const double PIover180  = 0.017453292519943295769139146242;
const double PIunder180 = 57.29577951308232087665461840231;
const double SqrtPI     = 1.772453850905516027298167483341;
const double SqrtPI2    = 2.506628274631000502415765284811;

const double Sqrt2      = 1.414213562373095048801688724210;
const double Sqrt3      = 1.732050807568877293527446341506;
const double Sqrt5      = 2.236067977499789696409173668731;
const double Sqrt6      = 2.449489742783178098197284074706;
const double Sqrt7      = 2.645751311064590590501615753639;
const double Sqrt12     = 3.464101615137754587054892683012;

const double Ln2        = 0.693147180559945309417232121458;
const double Ln10       = 2.302585092994045684017991454684;

inline int Trunc(const double a) {
	return a>=0? (int)a : (int)(a-1);
}
inline double Ceiling(const double a) {
	return a==(int)a? (int)a:
		a>0? (int)(1+(int)a): -(int)(1+(int)(-a));
}

inline double Rad(const double d)	{ return d*PIover180;  }
inline double Deg(const double r)	{ return r*PIunder180; }
inline double Cosd(const double a)	{ return cos(Rad(a));  }
inline double Sind(const double a)	{ return sin(Rad(a));  }
inline double Tand(const double a)	{ return tan(Rad(a));  }
//inline double Log(const double x)	{ return log(x)/Ln10;  }
//inline double Exp10(const double x)	{ return exp(x*Ln10);  }
//inline double Exp10f(const double x)	{ return expf(x*Ln10); }

/* --- Sign --- */
template <typename T>
inline int Sign(const T val) {
	return (T(0) < val) - (val < T(0));
}

/* --- Int --- */
template <typename T>
inline int Int(const T& x) {
	return x>=0? (int)x : (int)x - 1;
}

/* --- LongInt --- */
template <typename T>
inline long long LongInt(const T& x) {
	return x>=0? (long long)x : (long long)x - 1;
}

/* --- Round --- */
inline int Round(const double& a)
{
	return Int(a+0.5);
} /* Round */
//inline double Round(const double a) {
//	return a>=0? (int)(a+0.5): -(int)(0.5-a);
//}

/* --- LongRound --- */
inline long long LongRound(const double& a)
{
	return LongInt(a+0.5);
} /* LongRound */

/* --- Fractional --- */
template <typename T>
inline T Frac(const T& x) {
	return x - (T)Int(x);
}

/* --- RFractional --- */
template <typename T>
inline T RFrac(const T& x) {
	return 1.0 - (x - (T)Int(x));
}

/* --- Square --- */
template <typename T>
inline T Sq(const T& x) {
	return x*x;
}

/* --- Sqrt --- *
 * safe sqrt for negative numbers
 */
template <typename T>
inline T Sqrt(const T& x) {
	return x>0.0? sqrt(x): 0.0;
}

/* --- Cube --- */
template <typename T>
inline T Cube(const T& x) {
	return x*x*x;
}

/* --- Clamping --- */
#if __cplusplus >= 201703L
template<typename T>
const T& Clamp(const T& val, const T& low, const T& high)
{
	return std::clamp(val, low, high);
}
#else
template<typename T>
const T& Clamp(const T& val, const T& low, const T& high) {
	assert(low <= high);
	return (val<low)? low : (high<val) ? high : val;
}
#endif

#define PRINTDBL(val) do \
{\
	double _my_val = (val);\
	printf("%s -> %30.30f (%g) %llx\n", #val, _my_val, _my_val, *(unsigned long long *)&_my_val);\
} while (0)

#if defined(__APPLE__) || defined(ANDROID) || defined(__ANDROID_API__) || defined(CYGWIN)
inline void	sincos(const double x, double *ss, double *cc){ *ss = sin(x); *cc = cos(x);}
inline double	exp10(const double x)	{ return exp(x*Ln10);  }
inline float	exp10f(const float x)	{ return expf(x*Ln10); }
#endif

/*
  diff_of_products() computes a*b-c*d with a maximum error < 1.5 ulp

  Claude-Pierre Jeannerod, Nicolas Louvet, and Jean-Michel Muller,
  "Further Analysis of Kahan's Algorithm for the Accurate Computation
  of 2x2 Determinants". Mathematics of Computation, Vol. 82, No. 284,
  Oct. 2013, pp. 2245-2264
*/
inline double diff_of_products (double a, double b, double c, double d)
{
	double w = c * d;
	double e = fma(-d, c,  w);	/* fma(a,b,c) = a*b+c */
	double f = fma( a, b, -w);
	return f + e;
} // diff_of_products

///* compute the real roots of a quadratic equation: ax² + bx + c = 0,
//   provided the discriminant b²-4ac is positive
//*/
//inline void solve_quadratic (double a, double b, double c, double *x0, double *x1)
//{
//	double q = -0.5 * (b + copysign (sqrt (diff_of_products (b, b, 4.0*a, c)), b));
//	*x0 = q / a;
//	*x1 = c / q;
//} // solve_quadratic

/**
 * Numerically robust quadratic equation: x^2 + bx + c = 0.
 * @param a,b,c	coefficients
 * @param x1,x2	two real solutions.
 *              In case of no real solution then -b/2a is returned
 * @param acc	accuracy of comparison operations
 * @return the number of real solutions
 */
template <typename T>
inline int quadratic(const T b, const T c, T* x1, T* x2, const T acc)
{
	T b2 = b*b;
	T c4 = 4.0*c;
	T D  = b2 - c4;
	T prec = acc*Max(b2, Abs(c4));
	if (D <= prec) {
		*x1 = *x2 = -0.5*b;	// Always return this as a solution!!!
		if (D < -prec) return 0;
		return 1;
	} else {
		T bD;
		if (b>0.0)
			bD = -b - sqrt(D);
		else
			bD = -b + sqrt(D);
		*x1 = 0.5 * bD;
		*x2 = 2.0 * c / bD;	// Vieta's Formula
		return 2;
	}
} // quadratic

/**
 * Numerically robust quadratic equation: x^2 + bx + c = 0.
 * @param a,b,c	coefficients
 * @param x1,x2	two real solutions.
 *              In case of no real solution then -b/2a is returned
 * @param acc	accuracy of comparison operations
 * @return the number of real solutions
 */
inline int quadratic(const double b, const double c,
		double *x1, double *x2, const double acc)
{
	double D  = b*b - 4.0*c;
//	double D  = fma(b,b,-4.0*c);
	double prec = acc*Max(b, Abs(4*c));
	if (D <= prec) {
		*x1 = *x2 = -0.5*b;	// Always return this as a solution!!!
		if (D < -prec) return 0;
		return 1;
	} else {
		double bD;
		if (b>0.0)
			bD = -b - sqrt(D);
		else
			bD = -b + sqrt(D);
		*x1 = 0.5 * bD;
		*x2 = 2.0 * c / bD;	// Vieta's Formula
		return 2;
	}
} // quadratic

//inline int quadratic(const double a, const double b, const double c,
//		double *x1, double *x2, const double acc)
//{
////	double b2 = b*b;
//	double a4 = 4.0*a;
//	double D  = diff_of_products(b,b, a4, c);
//	double prec = acc*Max(b*b, Abs(a4*c));
//	if (D <= prec) {
//		*x1 = *x2 = -b/a/2.0;	// Always return this as a solution!!!
//		if (D < -prec) return 0;
//		return 1;
//	} else {
//		double bD;
//		if (b>0.0)
//			bD = (-b - sqrt(D)) / a;
//		else
//			bD = (-b + sqrt(D)) / a;
//		*x1 = 0.5 * bD;
//		*x2 = 2.0 * c / bD / a;	// Vieta's Formula
//		return 2;
//	}
//} // quadratic

/** calculate the zeros of the quadratic a*z**2+b*z+c.
 * the quadratic formula, modified to avoid
 * overflow, is used to find the larger zero if the
 * zeros are real and both zeros are complex.
 * the smaller real zero is found directly from the
 * product of the zeros c/a.
 */
inline void iquadratic(	double a, double b, double c,
			double* xr, double* xi,
			double* yr, double* yi)
{
	if (a==0.0) {
		*xr = (b!=0.0)? -c/b : 0.0;
		*yr = *xi = *yi = 0.0;
		return;
	}

	if (c==0.0) {
		*yr = -b/a;
		*xr = *xi = *yi = 0.0;
		return;
	}

	// compute discriminant avoiding overflow
	double b2 = b/2.;
	double e, d;
	if (Abs(b2)<Abs(c)) {
		e = (c<0.0)? -a : a;
		e = b2*(b2/Abs(c)) - e;
		d = sqrt(Abs(e))*sqrt(Abs(c));
	} else {
		e = 1.0 - (a/b2)*(c/b2);
		d = sqrt(Abs(e))*Abs(b2);
	}

	if (e>=0.0) {
		// real zeros
		if (b2>=0.0) d = -d;
		*yr = (-b2+d)/a;
		*xr = 0.0;
		if (*yr!=0.0) *xr = (c / *yr)/a;
		*xi = *yi = 0.0;
		return;
	} else {
		// complex conjugate zeros
		*xr = -b2/a;
		*yr = *xr;
		*xi = Abs(d/a);
		*yi = -*xi;
	}
} // iquadratic

/**
 * rounded sincos according to eps
 */
inline void bsincos(const double x, double *si, double *co)
{
	if (Abs(x)<2e-8) {
		*si = x;
		*co = 1 - x*x/2.0;
	} else
		sincos(x, si, co);
	if (*si==1 || *si==-1) *co = 0;
	else
	if (*co==1 || *co==-1) *si = 0;
} // bsincos

/* ================================= KahanSum =============================== *
 * Kahan sum as a class for easier implementation
 */
class KahanSum {
public:
	double sum;
	double c;
public:
	KahanSum(const double n=0.0) : sum(n), c(0.0) {}
	void	add(const double n) {
			double y = n - c;	// So far, so good: c is zero.
			double t = sum + y;	// Alas, sum is big, y small,
						// so low-order digits of y are lost.
			c = (t - sum) - y;	// (t - sum) recovers the
						// high-order part of y;
						// subtracting y recovers -(low part of y)
			sum = t;		// Algebraically, c should always be zero.
						// Beware eagerly optimising compilers!
			// Next time around, the lost low part will be added to y in a
			// fresh attempt.
		}
	KahanSum& operator += (const double n)	{ add(n); return *this; }
	void	operator()(const double n)	{ add(n); }
	double	operator()()		const	{ return sum; }
		operator double()	const	{ return sum; }
}; // KahanSum

/** Gamma FP error bounds according to
 * Higham, N. J. Accuracy and Stability of Numerical Algorithms (2nd ed.).
 * Philadelphia: Society for Industrial and Applied Mathematics.
 */
inline constexpr double fpgamma(int n)
{
	return (n * std::numeric_limits<double>::epsilon() * 0.5 )
		/ (1 - n * std::numeric_limits<double>::epsilon() * 0.5);
} // fpgamma

/* ---- Function prototypes ---- */
unsigned isqrt(unsigned n);
int	cubic(	const double a, const double b, const double c, double x[],
		const double acc, const int iter);
int	quartic(const double a, const double b, const double c, const double d,
		double x[], const double acc, const int iter);
int	polpol(const double C[], const int n, double *x, const double eps);
double	kahanSum(const int n, const double *input);
void	fpetrap(bool set=true);

#endif
