/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#ifndef __MATH_CONIC_H
#define __MATH_CONIC_H

#include <ostream>

#include "bmath.h"
#include "vector2d.h"

class Matrix3;

/**
 * Generalized conic line
 * S = ax^2 + 2hxy + by^2 + 2gx + 2fy + c = 0
 *
 *        /x\T  / a h g \   /x\
 *    S = |y| * | h b f | * |y| = 0
 *        \1/   \ g f c /   \1/
 *
 * Invariants
 *        | a h g |
 *    D = | h b f |
 *        | g f c |
 *
 *    D = a*b*c + 2*f*g*h - a*f*f - b*g*g - c*h*h
 *    I = a + b
 *    J = a*b - h^2
 *
 * Parameter to discriminate parallel lines real vs complex
 *    K = ca - g^2 + bc - f^2
 *
 * relative invariants
 *    I' = r*I
 *    J' = r^2*J
 *    D' = r^3*D
 *
 *
 *   Conic                                    D     J    D/I    K
 *   -------------------------------------- ----- ----- ----- -----
 *   Real ellipse                            !0    +ve   -ve    -
 *   Virtual ellipse                         !0    +ve   +ve    -
 *   Hyperbola                               !0    -ve    -     -
 *   Parabola                                !0     0     -     -
 *   Real intersecting lines                  0    -ve    -     -
 *   Conjugate complex intersecting lines     0    +ve    -     -
 *   Real distinct parallel lines             0     0     -    -ve
 *   Conjugate complex parallel lines         0     0     -    +ve
 *   Coincident lines                         0     0     -     0
 */
class Conic {
public:
	// Conic types
	enum class Type : int {
		DEGENERATE,
		POINT,
		LINE,
		LINES,
		ELLIPSE,
		PARABOLA,
		HYPERBOLA,
		//HYPERBOLA_HALF,
		UNKNOWN
	};

public:
	double a, h, b, g, f, c;
private:
	Type	_type;
	double	D, I, J, K;
	double	c1, c2, c3;
	double	c4, c5, c6;

static	const char* _typeStr[];

public:
static	const double CONICPREC;

public:
	Conic() { set(0.0, 0.0, 0.0, 0.0, 0.0, 0.0); }
	Conic(const double ag, const double af, const double ac) {
			set(0, 0, 0, ag, af, ac);
			assert(_type==Type::LINE);
		}
	Conic(const double aa, const double ah, const double ab,
	      const double ag, const double af, const double ac)
		{ set(aa, ah, ab, ag, af, ac); }

	Conic(const Conic&) = default;
	Conic& operator = (const Conic&) = default;

	void	set(const double aa, const double ah, const double ab,
		    const double ag, const double af, const double ac);
	void	set(const double ag, const double af, const double ac) {
			set(0, 0, 0, ag, af, ac);
			assert(_type==Type::LINE);
		}

	void	copy(const Conic& src);

	Type	type() const	{ return _type; }
	void	matrix(Matrix3* m) const;

	/** @return the rotation of axes to eliminate the h component */
	double	theta() const	{ return 0.5 * atan2(2.0*h, a - b); }
	void	rotate(const double theta);
	void	translate2Origin(double* dx, double* dy) const;
	void	translate(const double dx, const double dy);
	void	splitLines(Conic* conic1, Conic* conic2);
	void	mergeLines(const Conic& conic1, const Conic& conic2);

	int	intersect(const Conic& conic, Vector2D pts[4]) const;

	void	parametric();
	double	operator () (const double x, const double y) const
		{ return (a*x + 2.0*(h*y + g))*x + (b*y + 2.0*f)*y + c; }

	double	getT(double x, double y) const;
	void	getXY(const double t, double *x, double *y) const;
	int	getY(const double x, double *y1, double *y2) const;

	void	grad(const double x, const double y, double *dx, double *dy) const;
	double	adist(const double x, const double y) const;
	bool	equal(const Conic& conic, const double acc) const;

const	char*	typeStr()		const	{ return _typeStr[(int)_type]; }
static	const char* typeStr(Type t)		{ return _typeStr[(int)t]; }

private:
#ifdef STRONG
	int	_intersect(const Conic& conic, Vector2D pts[4]) const;
#endif
	int	_intersectLine(const Conic& line, Vector2D pts[4]) const;
	bool	addPoint(const Conic& conic, int *n, Vector2D pts[4], double x, double y) const;

friend	std::ostream& operator << (std::ostream&, const Conic&);
}; // Conic

std::ostream& operator << (std::ostream&, const Conic&);

#endif
