/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	14-Jun-2013
 */

#ifndef __MATH_MATRIX2_H
#define __MATH_MATRIX2_H

#include <ostream>
#include <string.h>

#include "vector.h"

// ================================== Matrix2 ==================================
/**
 * Matrix2 2x2 array
 * WARNING data are uninitialized on constructor
 */
class Matrix2 {
private:
	double	data[4];		// data
public:
static	constexpr double _identity[4] = {
			1.0, 0.0,
			0.0, 1.0,
		};

public:
	Matrix2()			{}
	Matrix2(int type)		{ if (type) identity(); else zero(); }
	Matrix2(const Matrix2& mat)	{ copy(mat); }
	Matrix2(std::initializer_list<double> mat) {
				assert(mat.size() == 4);
				std::copy(mat.begin(), mat.end(), data);
			}
	Matrix2& operator =(const Matrix2& matrix) { copy(matrix); return *this; }

	// Get/Set data
	int	rows()		const { return 2; }
	int	columns()	const { return 2; }
	int	cols()		const { return 2; }

	/** @return data at [row, col] */
	double	operator () (const int row, const int col) const {
			assert(row<rows() && col<columns());
			return data[row*columns()+col];
		}
	double&	operator () (const int row, const int col) {
			assert(row<rows() && col<columns());
			return data[row*columns()+col];
		}

	/** copy matrix from src
	 * @param src matrix to copy from
	 */
	void	copy(const Matrix2& src)	{ memcpy(data, src.data, sizeof(data)); }
	bool	isIdentity()	const	{ return !memcmp(data, _identity, sizeof(data)); }

	void	zero()		{ bzero(data, sizeof(data)); }
	void	identity()	{ memcpy(data, _identity, sizeof(data)); }
	void	transpose();
	void	T()		{ transpose(); }
	void	negate();

	Matrix2	operator *(const double f) const {
			Matrix2 mat;
			for (int i=0; i<rows(); i++)
				for (int j=0; j<cols(); j++)
					mat(i,j) = f*M(i,j);
			return mat;
		}

inline	void	multiply(const Matrix2& A, const Matrix2& B);
inline	void	multiplyT(const Matrix2& A, const Matrix2& B);
//	Matrix2& operator *= (const Matrix2&);

	bool	inverse(const double eps=Vector::_epsilon);
//	bool	inverse(const Matrix2& src) { invertMatrix(src.data, data);}

	double	trace() const	{ return M(0,0) + M(1,1); }

	Vector	operator *(const Vector &v) const {
			return Vector(v.x*M(0,0) + v.y*M(0,1),
				      v.x*M(1,0) + v.y*M(1,1),
				      v.z);
		}
	Vector	mult(const double x, const double y) const {
			return Vector(x*M(0,0) + y*M(0,1),
				      x*M(1,0) + y*M(1,1),
				      0.0);
		}
	Vector	mult(const Vector& v) const { return mult(v.x, v.y); }

	void	scale(const double sx, const double sy=0.0) {
			identity();
			M(0,0) = sx;
			M(1,1) = (sy==0.0?sx:sy);
		}

	void	rotate(const double angle);

	/** @return determinate */
	double	det()	const { return M(0,0)*M(1,1) - M(0,1)*M(1,0); }

	void	fix();

private:
	// same as operator() but for internal use to avoid all M
	// making code not readable
	double	M(const int row, const int col) const	{ return data[row*cols()+col]; }
	double&	M(const int row, const int col)		{ return data[row*cols()+col]; }
}; // Matrix2

std::ostream& operator << (std::ostream&, const Matrix2&);

/** multiply */
inline void Matrix2::multiply(const Matrix2& A, const Matrix2& B)
{
	M(0,0) = A(0,0)*B(0,0) + A(0,1)*B(1,0);
	M(0,1) = A(0,0)*B(0,1) + A(0,1)*B(1,1);

	M(1,0) = A(1,0)*B(0,0) + A(1,1)*B(1,0);
	M(1,1) = A(1,0)*B(0,1) + A(1,1)*B(1,1);
} /* multiply */

/** multiplyT
 * multiply the transpose of A with B: A.T() * B
 */
inline void Matrix2::multiplyT(const Matrix2& A, const Matrix2& B)
{
	M(0,0) = A(0,0)*B(0,0) + A(1,0)*B(1,0);
	M(0,1) = A(0,0)*B(0,1) + A(1,0)*B(1,1);

	M(1,0) = A(0,1)*B(0,0) + A(1,1)*B(1,0);
	M(1,1) = A(0,1)*B(0,1) + A(1,1)*B(1,1);
} /* multiplyT */

inline Matrix2 operator *(const double f, const Matrix2& mat)	{ return mat*f; }
#endif
