/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#ifndef __MATH_MATRIX3_H
#define __MATH_MATRIX3_H

#include <ostream>
#include <string.h>

#include "vector.h"

// ================================== Matrix3 ==================================
/**
 * Matrix3 3x3 array
 * WARNING data are uninitialized on constructor
 */
class Matrix3 {
private:
	double	data[9];		// data
public:
static	constexpr double _identity[9] = {
			1.0, 0.0, 0.0,
			0.0, 1.0, 0.0,
			0.0, 0.0, 1.0,
		};

public:
	Matrix3()			{}
	Matrix3(int type)		{ if (type) identity(); else zero(); }
	Matrix3(const Matrix3& mat)	{ copy(mat); }
	Matrix3(std::initializer_list<double> mat) {
				assert(mat.size() == 9);
				std::copy(mat.begin(), mat.end(), data);
			}
	Matrix3& operator =(const Matrix3& matrix) { copy(matrix); return *this; }

	// Get/Set data
	int	rows()		const { return 3; }
	int	columns()	const { return 3; }
	int	cols()		const { return 3; }

	/** @return data at [row, col] */
	double	operator () (const int row, const int col) const {
			assert(row<rows() && col<columns());
			return data[row*columns()+col];
		}
	double&	operator () (const int row, const int col) {
			assert(row<rows() && col<columns());
			return data[row*columns()+col];
		}

	/** copy matrix from src
	 * @param src matrix to copy from
	 */
	void	copy(const Matrix3& src)	{ memcpy(data, src.data, sizeof(data)); }
	bool	isIdentity()	const	{ return !memcmp(data, _identity, sizeof(data)); }

	void	zero()		{ bzero(data, sizeof(data)); }
	void	identity()	{ memcpy(data, _identity, sizeof(data)); }
	void	transpose();
	void	T()		{ transpose(); }
	void	negate();

	Matrix3	operator *(const double f) const {
			Matrix3 mat;
			for (int i=0; i<rows(); i++)
				for (int j=0; j<cols(); j++)
					mat(i,j) = f*M(i,j);
			return mat;
		}

inline	void	multiply(const Matrix3& A, const Matrix3& B);
inline	void	multiplyT(const Matrix3& A, const Matrix3& B);
//	Matrix3& operator *= (const Matrix3&);

	Vector	operator *(const Vector &v) const {
			return Vector(v.x*M(0,0) + v.y*M(0,1) + v.z*M(0,2),
				      v.x*M(1,0) + v.y*M(1,1) + v.z*M(1,2),
				      v.x*M(2,0) + v.y*M(2,1) + v.z*M(2,2));
		}
	Vector	mult(const double x, const double y, const double z) const {
			return Vector(x*M(0,0) + y*M(0,1) + z*M(0,2),
				      x*M(1,0) + y*M(1,1) + z*M(1,2),
				      x*M(2,0) + y*M(2,1) + z*M(2,2));
		}
	Vector	mult(const double x, const double y) const {
			return Vector(x*M(0,0) + y*M(0,1),
				      x*M(1,0) + y*M(1,1),
				      x*M(2,0) + y*M(2,1));
		}
	Vector	mult(const Vector& v) const { return mult(v.x, v.y, v.z); }

	Matrix3	operator +(const Matrix3& B) const {
			Matrix3 mat;
			for (int i=0; i<rows(); i++)
				for (int j=0; j<cols(); j++)
					mat(i,j) = M(i,j) + B(i,j);
			return mat;
		}

	Matrix3	operator -(const Matrix3& B) const {
			Matrix3 mat;
			for (int i=0; i<rows(); i++)
				for (int j=0; j<cols(); j++)
					mat(i,j) = M(i,j) - B(i,j);
			return mat;
		}

	bool	inverse(const double eps=Vector::_epsilon);
//	bool	inverse(const Matrix3& src) { invertMatrix(src.data, data);}

	/** @return trace of matrix (sum of diagonal elements) */
	double	trace() const	{ return M(0,0) + M(1,1) + M(2,2); }

	void	scale(const double sx, const double sy=0.0, const double sz=0.0) {
			identity();
			M(0,0) = sx;
			M(1,1) = (sy==0.0?sx:sy);
			M(2,2) = (sz==0.0?sx:sz);
		}

	void	rotate(const double angle, const int axis);
	void	rotate(const double angle, const double x, const double y,
				const double z);
	void	rotate(const double angle, const Vector& axis)
			{ rotate(angle, axis.x, axis.y, axis.z); }

	void	rotX(const double angle)	{rotate(angle,0);}
	void	rotY(const double angle)	{rotate(angle,1);}
	void	rotZ(const double angle)	{rotate(angle,2);}

	void	make(const Vector& vx, const Vector& vy, const Vector& vz) {
			identity();
			for (int i=0; i<cols(); i++) {
				M(0,i) = vx[i];
				M(1,i) = vy[i];
				M(2,i) = vz[i];
			}
		}

	/** @return determinate */
	double	det() const {
			return	  M(0,0)*M(1,1)*M(2,2)
				+ M(0,1)*M(1,2)*M(2,0)
				+ M(0,2)*M(1,0)*M(2,1)
				- M(0,2)*M(1,1)*M(2,0)
				- M(0,1)*M(1,0)*M(2,2)
				- M(0,0)*M(1,2)*M(2,1);
		}

	/** @return rank of matrix, non-zero rows after rref */
	int	rank(const double eps=Vector::_epsilon);

	/** @return eigen values of a symmetric matrix
	 *  @see https://en.wikipedia.org/wiki/Eigenvalue_algorithm
	 */
	void	eigenValuesOfSymmetric(double eig[3]) const;

private:
	// same as operator() but for internal use to avoid all M
	// making code not readable
	double	M(const int row, const int col) const	{ return data[row*cols()+col]; }
	double&	M(const int row, const int col)		{ return data[row*cols()+col]; }
}; // Matrix3

std::ostream& operator << (std::ostream&, const Matrix3&);

/** multiply */
inline void Matrix3::multiply(const Matrix3& A, const Matrix3& B)
{
	for (int i=0; i<rows(); i++)
		for (int j=0; j<cols(); j++)
			M(i,j) = A(i,0)*B(0,j) +
				 A(i,1)*B(1,j) +
				 A(i,2)*B(2,j);
} /* multiply */

/** multiplyT
 * multiply the transpose of A with B: A.T() * B
 */
inline void Matrix3::multiplyT(const Matrix3& A, const Matrix3& B)
{
	for (int i=0; i<rows(); i++)
		for (int j=0; j<cols(); j++)
			M(i,j) = A(0,i)*B(0,j) +
				 A(1,i)*B(1,j) +
				 A(2,i)*B(2,j);
} /* multiplyT */

inline Matrix3 operator *(const double f, const Matrix3& mat)	{ return mat*f; }

#endif
