/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#ifndef __MATH_MATRIX4_H
#define __MATH_MATRIX4_H

#include <ostream>
#include <assert.h>
#include <string.h>

#include "point.h"
#include "matrix3.h"

// ================================== Matrix4 ==================================
/**
 * Matrix4 4x4 array
 * WARNING data are uninitialized on constructor
 */
class Matrix4 {
private:
	double	data[16];		// data
public:
static	constexpr double _identity[16] = {
			1.0, 0.0, 0.0, 0.0,
			0.0, 1.0, 0.0, 0.0,
			0.0, 0.0, 1.0, 0.0,
			0.0, 0.0, 0.0, 1.0
		};

public:
	Matrix4()			{}
	Matrix4(int type)		{ if (type) identity(); else zero(); }
	Matrix4(const Matrix4& mat)	{ copy(mat); }
	Matrix4(std::initializer_list<double> mat) {
				assert(mat.size() == 16);
				std::copy(mat.begin(), mat.end(), data);
			}
	Matrix4& operator =(const Matrix4& matrix) { copy(matrix); return *this; }

	// Get/Set data
	int	rows()		const { return 4; }
	int	columns()	const { return 4; }
	int	cols()		const { return 4; }

	/** @return data at [row, col] */
	double	operator () (const int row, const int col) const {
			assert(row<rows() && col<columns());
			return data[row*columns()+col];
		}
	double&	operator () (const int row, const int col) {
			assert(row<rows() && col<columns());
			return data[row*columns()+col];
		}

	/** copy matrix from src
	 * @param src matrix to copy from
	 */
	void	copy(const Matrix4& src) { memcpy(data, src.data, sizeof(data)); }
	bool	isIdentity()	const	 { return !memcmp(data, _identity, sizeof(data)); }

	void	zero()		{ bzero(data, sizeof(data)); }
	void	identity()	{ assert(_identity[0]==1); memcpy(data, _identity, sizeof(data)); }
	void	transpose();
	void	T()		{ transpose(); }
	void	negate();

	/** set the 3x3 matrix as the rotations only */
	void	set(const Matrix3& src) {
			for (int j=0; j<src.rows(); j++)
				for (int i=0; i<src.columns(); i++)
					M(j,i) = src(j,i);
		}

	/** @return transformation matrix basis vectors */
	Vector	basisX()	const	{ return Vector(M(0,0), M(0,1), M(0,2)); }
	Vector	basisY()	const	{ return Vector(M(1,0), M(1,1), M(1,2)); }
	Vector	basisZ()	const	{ return Vector(M(2,0), M(2,1), M(2,2)); }

	/** set basis vectors */
	void	basisX(const Vector& v)	{ M(0,0)=v.x; M(0,1)=v.y; M(0,2)=v.z; }
	void	basisY(const Vector& v)	{ M(1,0)=v.x; M(1,1)=v.y; M(1,2)=v.z; }
	void	basisZ(const Vector& v)	{ M(2,0)=v.x; M(2,1)=v.y; M(2,2)=v.z; }

	Matrix4	operator *(const double f) const {
			Matrix4 mat;
			for (int i=0; i<rows(); i++)
				for (int j=0; j<cols(); j++)
					mat(i,j) = f*M(i,j);
			return mat;
		}

inline	void	multiply(const Matrix4& A, const Matrix4& B);
inline	void	multiplyT(const Matrix4& A, const Matrix4& B);
	Matrix4 operator *(const Matrix4& B)	{ Matrix4 r; r.multiply(*this, B); return r; }
	Matrix4& operator *=(const Matrix4& B)	{ Matrix4 r; r.multiply(*this, B); copy(r); return *this; }

	void	inverse();
	void	inverse(const Matrix4& src) { invertMatrix(src.data, data);}
	double	trace() const;

	/** multiple with a point (x,y,z) */
	Point	mult(const double x, const double y, const double z) const {
			return Vector(x*M(0,0) + y*M(0,1) + z*M(0,2) + M(0,3),
				      x*M(1,0) + y*M(1,1) + z*M(1,2) + M(1,3),
				      x*M(2,0) + y*M(2,1) + z*M(2,2) + M(2,3));
		}
	/** multiple with a Point (x,y,0) */
	Point	mult(const double x, const double y) const {
			return Vector(x*M(0,0) + y*M(0,1) + M(0,3),
				      x*M(1,0) + y*M(1,1) + M(1,3),
				      x*M(2,0) + y*M(2,1) + M(2,3));
		}
	/** multiple with a Point p */
	Point	operator *(const Point &p) const { return mult(p.x, p.y, p.z); }

	/** multiply with a Vector (x,y,z) [TRANSLATION NOT INCLUDED] */
	Vector	multVector(const double x, const double y, const double z) const {
			return Vector(x*M(0,0) + y*M(0,1) + z*M(0,2),
				      x*M(1,0) + y*M(1,1) + z*M(1,2),
				      x*M(2,0) + y*M(2,1) + z*M(2,2));
		}
	/** multiply with a Vector (x,y,0) [TRANSLATION NOT INCLUDED] */
	Vector	multVector(const double x, const double y) const {
			return Vector(x*M(0,0) + y*M(0,1),
				      x*M(1,0) + y*M(1,1),
				      x*M(2,0) + y*M(2,1));
		}
	/** multiply with a Vector v operator [TRANSLATION NOT INCLUDED] */
	Vector	operator *(const Vector &v) const { return multVector(v.x, v.y, v.z); }

	/** multiply the TRANSPOSE matrix with a vector */
	Vector	multT(const Vector& v) const {
			return Vector(v.x*M(0,0) + v.y*M(1,0) + v.z*M(2,0),
				      v.x*M(0,1) + v.y*M(1,1) + v.z*M(2,1),
				      v.x*M(0,2) + v.y*M(1,2) + v.z*M(2,2));
		}

	/** transform Point in place (x,y,z) = Matrix * (x,y,z) */
	void	transform(double* x, double* y, double* z) const {
			double ox = *x*M(0,0) + *y*M(0,1) + *z*M(0,2) + M(0,3);
			double oy = *x*M(1,0) + *y*M(1,1) + *z*M(1,2) + M(1,3);
			double oz = *x*M(2,0) + *y*M(2,1) + *z*M(2,2) + M(2,3);
			*x = ox;
			*y = oy;
			*z = oz;
		}
	/** transform Point p in place p = Matrix * p */
	void	transform(Point& p) const	{ transform(&p.x, &p.y, &p.z); }

	/** transform Vector in place (x,y,z) = Matrix * (x,y,z) [TRANSLATION NOT INCLUDED] */
	void	transformVector(double* x, double* y, double* z) const {
			double ox = *x*M(0,0) + *y*M(0,1) + *z*M(0,2);
			double oy = *x*M(1,0) + *y*M(1,1) + *z*M(1,2);
			double oz = *x*M(2,0) + *y*M(2,1) + *z*M(2,2);
			*x = ox;
			*y = oy;
			*z = oz;
		}
	/** transform Vector in place v = Matrix * v [TRANSLATION NOT INCLUDED] */
	void	transform(Vector& v) const	{ transformVector(&v.x, &v.y, &v.z); }

	/** Create translation matrix */
	void	translate(const double x, const double y, const double z) {
			identity();
			M(0,3) = x;
			M(1,3) = y;
			M(2,3) = z;
		}
	void	translate(const Vector& v) { translate(v.x, v.y, v.z); }

	void	scale(const double sx, const double sy=0.0, const double sz=0.0) {
			identity();
			M(0,0) = sx;
			M(1,1) = (sy==0.0?sx:sy);
			M(2,2) = (sz==0.0?sx:sz);
		}

	// Create rotation matrix
	void	rotate(const double angle, const int axis);
	void	rotate(const double angle, const double x, const double y, const double z);
	void	rotate(const double angle, const Vector& axis)
			{ rotate(angle, axis.x, axis.y, axis.z); }

	void	rotateX(const double angle)	{rotate(angle,0);}
	void	rotateY(const double angle)	{rotate(angle,1);}
	void	rotateZ(const double angle)	{rotate(angle,2);}

	void	make(const Vector& vx, const Vector& vy, const Vector& vz) {
			identity();
			for (int i=0; i<3; i++) {
				M(0,i) = vx[i];
				M(1,i) = vy[i];
				M(2,i) = vz[i];
			}
		}

	// set the translation component
	void	setT(const Vector& t) {
			M(0,3) = t[0];
			M(1,3) = t[1];
			M(2,3) = t[2];
		}

	/** @return determinate of the rotation 3x3 matrix only */
	double	det3() const {
			return	  M(0,0)*M(1,1)*M(2,2)
				+ M(0,1)*M(1,2)*M(2,0)
				+ M(0,2)*M(1,0)*M(2,1)
				- M(0,2)*M(1,1)*M(2,0)
				- M(0,1)*M(1,0)*M(2,2)
				- M(0,0)*M(1,2)*M(2,1);
		}

	/** @return determinate */
	double det() const;

	/** @return rank of matrix, non-zero rows after rref */
	int	rank(const double eps=Vector::_epsilon);

	/** try to correct precision errors from transformation matrices */
	void	fix(bool rotation=false);

private:
	// same as operator() but for internal use to avoid all M
	// making code not readable
	double	M(const int row, const int col) const	{ return data[row*cols()+col]; }
	double&	M(const int row, const int col)		{ return data[row*cols()+col]; }

	void	fix01();
static	void	invertMatrixGeneral(const double*, double*);
static	void	invertMatrix(const double*, double*);
}; // Matrix4

std::ostream& operator << (std::ostream&, const Matrix4&);

/** multiply */
inline void Matrix4::multiply(const Matrix4& A, const Matrix4& B)
{
	for (int i=0; i<rows(); i++)
		for (int j=0; j<cols(); j++)
			M(i,j) =	A(i,0)*B(0,j) +
					A(i,1)*B(1,j) +
					A(i,2)*B(2,j) +
					A(i,3)*B(3,j);
} /* multiply */

/** multiplyT
 * multiply the transpose of A with B: A.T() * B
 */
inline void Matrix4::multiplyT(const Matrix4& A, const Matrix4& B)
{
	for (int i=0; i<rows(); i++)
		for (int j=0; j<cols(); j++)
			M(i,j) = A(0,i)*B(0,j) +
				 A(1,i)*B(1,j) +
				 A(2,i)*B(2,j) +
				 A(3,i)*B(3,j);
} /* multiplyT */

inline Matrix4 operator *(const double f, const Matrix4& mat)	{ return mat*f; }
#endif
