/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	11-Jan-2006
 */

#ifndef __TOOL_STREAM_H
#define __TOOL_STREAM_H

#include <string>
#include <fstream>

#include "os.h"

/**
 * Geometry streaming I/O class
 */
class Stream {
public:
	enum class IO {
		Read,
		Write
	};

protected:
	std::string	filename;	/** filename to open	*/
	std::fstream	stream;		/** Input stream	*/
	IO		_io;		/** I/O type		*/
	bool		binary;		/** binary file		*/
	bool		header;		/** is header written	*/
	bool		footer;		/** is footer written	*/

public:
	Stream() : _io(IO::Read) {}
	Stream(const std::string& name) : filename(name), _io(IO::Read) {}
	Stream(const std::string& name, IO io=IO::Read, bool bin=false) { open(name, io, bin); }
	~Stream()	{ close(); }

	/** open a new file for output */
	bool	open(IO io=IO::Read, bool bin=false);
	bool	open(const std::string& name, IO io=IO::Read, bool bin=false);

	/** close file and check headers if are needed */
	void	close();

	/** @return true is file is open for read */
	bool	isRead() const	{ return _io == IO::Read; }
	/** @return true if opened for writing */
	bool	isWrite() const	{ return _io == IO::Write; }

	/** @return true is file is open */
	bool	isOpen()		{ return stream.is_open(); }

	/** skip entire line */
	void	skipLine();

	/** write a header before any other write statement
	 * @see close
	 * @see writeFooter
	 */
virtual	void	writeHeader();

	/** write a footer before closing the file
	 * @see close
	 * @see writeHeader
	 */
virtual	void	writeFooter();

	/** test if header is needed
	 * @see writeHeader
	 */
virtual	void	write();
}; // Stream

#endif
