/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 *
 * Changes:
 * - Added DXF export - 22.06.2010 C. Theis
 */

#include "viewer.h"
#ifdef STAT
#	include "tool/timer.h"
#endif

#define ERRORMSG	"Errors found"

using std::cout;
using std::cerr;
using std::endl;

static const char* projectionMessage[] = {
		/*  0 */ "Nothing",
		/*  1 */ "Spawn",
		/*  2 */ "Starting",
		/*  3 */ "Creating conics",
		/*  4 */ "Finding bodies and region location",
		/*  5 */ "Intersecting bodies",
		/*  6 */ "Scanning segments",
		/*  7 */ "Drawing",
		/*  8 */ "",
		/*  9 */ "",
		/* 10 */ "Finished"
	};

/** GeometryViewer */
GeometryViewer::GeometryViewer(const Geometry& g, GeometryKernel& k) :
	geometry(g),
	kernel(k),
	decoration(g, k, *this),
	palette(g, k, *this),

	d2(g, k, *this),
	d3(g, k, *this),

	lattice(g, k, *this),
	voxel(g,   k, *this),
	userdump(g,k, *this),
	usrbin(g,  k, *this),
	image(g,   k, *this),

	exporter(g, k, *this)
{
	title("Viewer");
	painter.init(640,480);	// some initial guess

	showErrors  = true;

	textBackgroundLevel = 200;

	// thread
	pthread_mutex_init(&mutex,     NULL);
	pthread_mutex_init(&mutexStop, NULL);
	pthread_mutex_init(&mutexDraw, NULL);

	thread    = 0;
	pStop     = false;
	pState    = Projection::NOTHING;
	_percent  = 0;
	endThread = NULL;
	_mask     = -1;
	_fast     = false;

	backgroundColor(geometry.backgroundColor());

	font.load(DEFAULT_FONT);
} // GeometryViewer

/** ~GeometryViewer */
GeometryViewer::~GeometryViewer()
{
	stopThread();
	pthread_mutex_destroy(&mutex);
	pthread_mutex_destroy(&mutexStop);
	pthread_mutex_destroy(&mutexDraw);
} /* ~GeometryViewer */

#if 0
/** makeLatticeColor */
void GeometryViewer::makeLatticeColor()
{
	latticeHashColor = Color32::darker(geometry.latticeColor, latticeLevel);
	voxelHashColor   = Color32::darker(geometry.voxelColor,   latticeLevel);
} // makeLatticeColor
#endif

/** backgroundColor */
void GeometryViewer::backgroundColor(const dword color)
{
	painter.background(color);
} // backgroundColor

/** origin */
void GeometryViewer::origin(const double x, const double y, const double z)
{
	stopThread();
	kernel.origin(x,y,z);
	image.matrix();
} // origin

/** moveViewOriginTo0
 * Move view origin to center of "view"
 */
void GeometryViewer::moveViewOriginTo0()
{
	stopThread();
	kernel.moveViewOriginTo0();
	image.matrix();
} // moveViewOriginTo0

/** matrix */
void GeometryViewer::matrix(const Matrix3& m)
{
	stopThread();				// Stop thread before modifying the matrix
	Matrix4 mat4 = kernel.view.matrix();	// get the matrix from the viewport
	mat4.set(m);				// copy the 3x3 rotations in the to 4x4 keeping the translation
	kernel.matrix(mat4);
	image.matrix();
} // matrix

/** matrix */
void GeometryViewer::matrix(const Matrix4& m)
{
	stopThread();	// Stop thread before modifying the matrix
	kernel.matrix(m);
	image.matrix();
} // matrix

/** resize */
void GeometryViewer::resize(int w, int h)
{
	if (pState == Projection::DRAW) stopThread();
	lock();
	lockDraw();
	painter.init(w, h);
	kernel.resize(w,h);
	unlockDraw();
	unlock();
} // resize

/** set projection state and percentage done
 * @return if thread should stop
 */
bool GeometryViewer::state(Projection s, int p)
{
	bool ps;

	ps = pStop;
	if (ps)
		_percent = 0;
	else {
		pState   = s;
		_percent = p;
	}
	return ps;
} // state

/* run project as a thread */
int GeometryViewer::spawnProject(NotifyFunc func, void* arg)
{
	if (pState == Projection::SPAWN) return 0;
	lock();
	DUMP(cout << "GeometryViewer::spawnProject(" << title() << ").stopThread()" << endl);
	stopThread();
	endThread = func;
	endArg    = arg;
	pState    = Projection::SPAWN;
	assert(thread==0);
	assert(!kernel.isRunning());
	int rc = pthread_create(&thread, NULL, (void*(*)(void*))GeometryViewer::runProject, this);
	if (rc) {
		thread = 0;
		perror("pthread_create");
		d2.project();
		kernel.error("System error spawning in background the projection.\n"
				"Switching to synchronous mode");
#ifdef STRONG
		cerr << "System error spawning in background the projection." << endl;
		cerr << "Switching to synchronous mode" << endl;
#endif
	}
	unlock();
	return rc;
} // spawnProject

/* static entry to run a project */
void* GeometryViewer::runProject(GeometryViewer* gthis)
{
	DUMP(cout << "*-* GeometryViewer::runProject(" << gthis->title() << ")" << endl);
	if (gthis->pStop) return NULL;
	if (gthis->pState != Projection::SPAWN) return NULL;

	gthis->d2.project();

	if (gthis->endThread) gthis->endThread(gthis->endArg);
	gthis->thread = 0;
	return NULL;
} // runProject

/* run draw as a thread */
int GeometryViewer::spawnDraw(NotifyFunc func, void* arg, int mask, bool fast)
{
	if (pState != Projection::FINISHED) return 0;
	lock();
	DUMP(cout << "GeometryViewer::spawnDraw(" << title() << ").stopThread()" << endl);
	stopThread();
	endThread = func;
	endArg    = arg;
	_mask     = mask;
	_fast     = fast;
	saveState();
	pState    = Projection::DRAW;
	assert(thread==0);
	assert(!kernel.isRunning());
	int rc = pthread_create(&thread, NULL, (void*(*)(void*))GeometryViewer::runDraw, this);
	if (rc) {
		thread = 0;
		perror("pthread_create");
		draw(mask, fast);
		kernel.error("System error spawning in background the late drawing.\n"
				"Switching to synchronous mode");
#ifdef STRONG
		cerr << "System error spawning in background the late drawing." << endl;
		cerr << "Switching to synchronous mode" << endl;
#endif
	}
	unlock();
	return rc;
} // spawnDraw

/* static entry to submit late drawing */
void* GeometryViewer::runDraw(GeometryViewer* gthis)
{
	if (gthis->pStop) return NULL;
	DUMP(cout << "*-* GeometryViewer::runDraw(" << gthis->title() << ")" << endl);
	assert(!gthis->kernel.isRunning());
	gthis->draw(gthis->_mask, gthis->_fast);
	assert(!gthis->kernel.isRunning());
	gthis->restoreState();
	gthis->thread = 0;
	if (gthis->endThread) gthis->endThread(gthis->endArg);
	return NULL;
} // runDraw

/* stop projection/drawing thread */
void GeometryViewer::stopThread()
{
	DUMP(cout << "*-* GeometryViewer::stopThread(" << title() << ")" << endl);
	lockStop();
	endThread = NULL;		 /* do not send end project cmd */
	pStop     = true;
	kernel.stopThread();
	if (thread) pthread_join(thread, NULL);
	thread = 0;
	pStop  = false;

	// stopThread has to finish with state either
	//  1. FINISHED
	//  2. NOTHING
	if (pState==Projection::DRAW || pState==Projection::FINISHED)
		pState = Projection::FINISHED;
	else {
		_percent = 0;
		pState   = Projection::NOTHING;
		d2.setInvalid();
	}
	assert(!kernel.isRunning());
	unlockStop();
} // stopThread

/** draw
 * @param mask	of options to draw
 * @return	projection state that was drawn
 */
GeometryViewer::Projection GeometryViewer::draw(int mask, bool fast)
{
#ifdef STAT
	Timer timer;
	timer.start();
#endif
	lockDraw();
	int  line = 0;
	Projection _state = pState;		// Remember projection state

	DUMP(cout << "*-* GeometryViewer::draw(" << title()
		<< ", " << mask << ", " << fast << ")" << endl);

	kernel.errorReset();

	bool show3D = d3.show and mask&DRAW_3D;
	if (!show3D) painter.frustum(-ZPainter::infinite, ZPainter::infinite);

	// Clean display
	if (mask & DRAW_CLEAR) painter.clear();

	kernel.updateBodyShow();
	if (_state>=Projection::SCAN and mask&DRAW_SEGMENTS && d2.show)
		d2.draw(painter);

	// fake percentage during spawn draw only!!!
	if (pState==Projection::DRAW) _percent = 10;

	if (_state==Projection::FINISHED || _state==Projection::DRAW || _state==Projection::NOTHING) {
		if (mask&DRAW_REGIONS && d2.fillRegions)	 d2.drawRegions(painter);
		decoration.draw(painter, mask);
		if (fast) {
			if (mask&DRAW_LATTICES   && lattice.show==SHOW_PROMPT)	lattice.draw(painter);
			if (mask&DRAW_VOXEL      && voxel.show==SHOW_PROMPT)	voxel.draw(painter);
			if (mask&DRAW_USRBIN     && usrbin.show==SHOW_PROMPT)	usrbin.draw(painter);
			if (mask&DRAW_IMAGE      && image.show==SHOW_PROMPT)	image.draw(painter);
		} else {
			// Warning on the priority order
			if (mask&DRAW_LATTICES   && lattice.show)	lattice.draw(painter);
			if (mask&DRAW_VOXEL      && voxel.show)		voxel.draw(painter);
			if (mask&DRAW_USRBIN     && usrbin.show)	usrbin.draw(painter);
			if (show3D)					d3.draw(painter);
			if (mask&DRAW_USERDUMP   && userdump.show)	userdump.draw(painter, show3D);
			if (mask&DRAW_IMAGE      && image.show)		image.draw(painter);
			if (mask&DRAW_REGIONS    && d2.fillRegions)	d2.drawRegionsLate(painter);
			if (mask&DRAW_LABELS     && d2.showLabels)	d2.drawLabels(painter);
			if (mask&DRAW_PALETTE    && palette.show)	palette.draw(painter);
			// run beauty passes for 2D layers
			if (d2.show) d2.postDraw(painter);
		}
	} else {
		decoration.draw(painter, mask);
		if (!fast) {
			decoration.drawProgress(painter, percent(), projectionMessage[(int)_state]);
			line++;
		}
	}
	if (pState==Projection::DRAW) _percent = 30;

	// Special 3D real-time drawing
	if (fast && (_state==Projection::FINISHED || _state==Projection::NOTHING)) {
		// WARNING should not be called during the SPAWN projection is ON
		if (show3D)	d3.draw(painter, true);
	}
	if (pState==Projection::DRAW) _percent = 70;

	if (_state==Projection::FINISHED or _state==Projection::DRAW) {
		// Draw 3D wireframe of bodies
		if (mask&DRAW_WIREFRAME and
		    kernel.bodyShow() & BIT_WIREFRAME) d3.drawWireframe(painter);
		// Draw bounding boxes
		if (mask&DRAW_BBOX and not fast) {
			if (kernel.bodyShow() & BIT_BBOX) d3.drawBodiesBBox(painter);
			d3.drawRegionsBBox(painter);
		}
	}
	if (pState==Projection::DRAW) _percent = 80;

	if (_state>=Projection::SCAN && mask&DRAW_VERTICES && d2.showVertex && d2.show)
		d2.drawVertices(painter);

	// Independent from the state
	//decoration.draw(painter, mask);

	// implement special post-draw effects (e.g. adaptive anti-aliasing)
	// that can be done after the beauty pass in the background
	if (_state == Projection::DRAW and show3D)
		d3.postDraw(painter);

	if (pState==Projection::DRAW) _percent = 90;

	if (mask&DRAW_STATUS && kernel.errors() && showErrors)
		decoration.drawMessage(painter, ERRORMSG, geometry.errorColor, line++);

	if (mask&DRAW_MESSAGE && !geometry.message().empty())
		decoration.drawMessage(painter, geometry.message().c_str(),
					geometry.messageColor, line++);

	if (pState==Projection::DRAW) _percent = 100;
	unlockDraw();
#ifdef STAT
	timer.stop();
	cerr << "Time to draw: " << timer << endl;
#endif
	return _state;
} // draw

/** memory */
size_t GeometryViewer::memory() const
{
	size_t mem = sizeof(GeometryViewer)
		+ kernel.memory()
		//+ voxel.memory()
		+ painter.memory();
	return mem;
} // memory

/** printMemory */
void GeometryViewer::printMemory() const
{
	cout << endl << "GeometryViewer:" << endl;
	cout << "Memory:"	<< endl;
	cout << "\tSelf:\t"	<< sizeof(GeometryViewer) << endl;
	cout << "\tVideo:\t"	<< painter.memory() << endl;

	cout << "\tKernel:\t"	<< kernel.memory() << endl;
	cout << "\tTotal:\t"	<< memory() << endl;
} // printMemory

